
import google.protobuf
from packaging.version import Version
if Version(google.protobuf.__version__).major >= 5:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: unity_catalog_prompt_service.proto
  # Protobuf Python Version: 5.26.0
  """Generated protocol buffer code."""
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import symbol_database as _symbol_database
  from google.protobuf.internal import builder as _builder
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()


  from . import databricks_pb2 as databricks__pb2
  from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
  from . import unity_catalog_prompt_messages_pb2 as unity_catalog_prompt_messages_pb2
  from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"unity_catalog_prompt_service.proto\x12\x13mlflow.unitycatalog\x1a\x10\x64\x61tabricks.proto\x1a\x15scalapb/scalapb.proto\x1a#unity_catalog_prompt_messages.proto\x1a\x1bgoogle/protobuf/empty.proto2\xfe \n\x19UnityCatalogPromptService\x12\xa9\x01\n\x0c\x43reatePrompt\x12(.mlflow.unitycatalog.CreatePromptRequest\x1a\x1b.mlflow.unitycatalog.Prompt\"R\xf2\x86\x19N\n+\n\x04POST\x12\x1d/mlflow/unity-catalog/prompts\x1a\x04\x08\x02\x10\x00\x10\x03*\x1d(Unity Catalog) Create Prompt\x12\xb1\x01\n\x0cUpdatePrompt\x12(.mlflow.unitycatalog.UpdatePromptRequest\x1a\x1b.mlflow.unitycatalog.Prompt\"Z\xf2\x86\x19V\n3\n\x05PATCH\x12$/mlflow/unity-catalog/prompts/{name}\x1a\x04\x08\x02\x10\x00\x10\x03*\x1d(Unity Catalog) Update Prompt\x12\xad\x01\n\x0c\x44\x65letePrompt\x12(.mlflow.unitycatalog.DeletePromptRequest\x1a\x16.google.protobuf.Empty\"[\xf2\x86\x19W\n4\n\x06\x44\x45LETE\x12$/mlflow/unity-catalog/prompts/{name}\x1a\x04\x08\x02\x10\x00\x10\x03*\x1d(Unity Catalog) Delete Prompt\x12\xa6\x01\n\tGetPrompt\x12%.mlflow.unitycatalog.GetPromptRequest\x1a\x1b.mlflow.unitycatalog.Prompt\"U\xf2\x86\x19Q\n1\n\x03GET\x12$/mlflow/unity-catalog/prompts/{name}\x1a\x04\x08\x02\x10\x00\x10\x03*\x1a(Unity Catalog) Get Prompt\x12\xc2\x01\n\rSearchPrompts\x12).mlflow.unitycatalog.SearchPromptsRequest\x1a*.mlflow.unitycatalog.SearchPromptsResponse\"Z\xf2\x86\x19V\n2\n\x04POST\x12$/mlflow/unity-catalog/prompts/search\x1a\x04\x08\x02\x10\x00\x10\x03*\x1e(Unity Catalog) Search Prompts\x12\xd6\x01\n\x13\x43reatePromptVersion\x12/.mlflow.unitycatalog.CreatePromptVersionRequest\x1a\".mlflow.unitycatalog.PromptVersion\"j\xf2\x86\x19\x66\n;\n\x04POST\x12-/mlflow/unity-catalog/prompts/{name}/versions\x1a\x04\x08\x02\x10\x00\x10\x03*%(Unity Catalog) Create Prompt Version\x12\xe1\x01\n\x13UpdatePromptVersion\x12/.mlflow.unitycatalog.UpdatePromptVersionRequest\x1a\".mlflow.unitycatalog.PromptVersion\"u\xf2\x86\x19q\nF\n\x05PATCH\x12\x37/mlflow/unity-catalog/prompts/{name}/versions/{version}\x1a\x04\x08\x02\x10\x00\x10\x03*%(Unity Catalog) Update Prompt Version\x12\xd6\x01\n\x13\x44\x65letePromptVersion\x12/.mlflow.unitycatalog.DeletePromptVersionRequest\x1a\x16.google.protobuf.Empty\"v\xf2\x86\x19r\nG\n\x06\x44\x45LETE\x12\x37/mlflow/unity-catalog/prompts/{name}/versions/{version}\x1a\x04\x08\x02\x10\x00\x10\x03*%(Unity Catalog) Delete Prompt Version\x12\xd6\x01\n\x10GetPromptVersion\x12,.mlflow.unitycatalog.GetPromptVersionRequest\x1a\".mlflow.unitycatalog.PromptVersion\"p\xf2\x86\x19l\nD\n\x03GET\x12\x37/mlflow/unity-catalog/prompts/{name}/versions/{version}\x1a\x04\x08\x02\x10\x00\x10\x03*\"(Unity Catalog) Get Prompt Version\x12\xfa\x01\n\x14SearchPromptVersions\x12\x30.mlflow.unitycatalog.SearchPromptVersionsRequest\x1a\x31.mlflow.unitycatalog.SearchPromptVersionsResponse\"}\xf2\x86\x19y\nB\n\x04POST\x12\x34/mlflow/unity-catalog/prompts/{name}/versions/search\x1a\x04\x08\x02\x10\x00\x10\x03\x18\xe8\x07\x18\xec\x07\x18\xba\x17\x18\x01*&(Unity Catalog) Search Prompt Versions\x12\xc2\x01\n\x0eSetPromptAlias\x12*.mlflow.unitycatalog.SetPromptAliasRequest\x1a\x16.google.protobuf.Empty\"l\xf2\x86\x19h\nB\n\x04POST\x12\x34/mlflow/unity-catalog/prompts/{name}/aliases/{alias}\x1a\x04\x08\x02\x10\x00\x10\x03* (Unity Catalog) Set Prompt Alias\x12\xcd\x01\n\x11\x44\x65letePromptAlias\x12-.mlflow.unitycatalog.DeletePromptAliasRequest\x1a\x16.google.protobuf.Empty\"q\xf2\x86\x19m\nD\n\x06\x44\x45LETE\x12\x34/mlflow/unity-catalog/prompts/{name}/aliases/{alias}\x1a\x04\x08\x02\x10\x00\x10\x03*#(Unity Catalog) Delete Prompt Alias\x12\xf5\x01\n\x17GetPromptVersionByAlias\x12\x33.mlflow.unitycatalog.GetPromptVersionByAliasRequest\x1a\".mlflow.unitycatalog.PromptVersion\"\x80\x01\xf2\x86\x19|\nK\n\x03GET\x12>/mlflow/unity-catalog/prompts/{name}/versions/by-alias/{alias}\x1a\x04\x08\x02\x10\x00\x10\x03*+(Unity Catalog) Get Prompt Version By Alias\x12\xb1\x01\n\x0cSetPromptTag\x12(.mlflow.unitycatalog.SetPromptTagRequest\x1a\x16.google.protobuf.Empty\"_\xf2\x86\x19[\n7\n\x04POST\x12)/mlflow/unity-catalog/prompts/{name}/tags\x1a\x04\x08\x02\x10\x00\x10\x03*\x1e(Unity Catalog) Set Prompt Tag\x12\xc2\x01\n\x0f\x44\x65letePromptTag\x12+.mlflow.unitycatalog.DeletePromptTagRequest\x1a\x16.google.protobuf.Empty\"j\xf2\x86\x19\x66\n?\n\x06\x44\x45LETE\x12//mlflow/unity-catalog/prompts/{name}/tags/{key}\x1a\x04\x08\x02\x10\x00\x10\x03*!(Unity Catalog) Delete Prompt Tag\x12\xda\x01\n\x13SetPromptVersionTag\x12/.mlflow.unitycatalog.SetPromptVersionTagRequest\x1a\x16.google.protobuf.Empty\"z\xf2\x86\x19v\nJ\n\x04POST\x12</mlflow/unity-catalog/prompts/{name}/versions/{version}/tags\x1a\x04\x08\x02\x10\x00\x10\x03*&(Unity Catalog) Set Prompt Version Tag\x12\xed\x01\n\x16\x44\x65letePromptVersionTag\x12\x32.mlflow.unitycatalog.DeletePromptVersionTagRequest\x1a\x16.google.protobuf.Empty\"\x86\x01\xf2\x86\x19\x81\x01\nR\n\x06\x44\x45LETE\x12\x42/mlflow/unity-catalog/prompts/{name}/versions/{version}/tags/{key}\x1a\x04\x08\x02\x10\x00\x10\x03*)(Unity Catalog) Delete Prompt Version Tag\x12\xe9\x01\n\x1aLinkPromptVersionsToModels\x12\x36.mlflow.unitycatalog.LinkPromptVersionsToModelsRequest\x1a\x16.google.protobuf.Empty\"{\xf2\x86\x19w\nC\n\x04POST\x12\x35/mlflow/unity-catalog/prompt-versions/links-to-models\x1a\x04\x08\x02\x10\x00\x10\x03*.(Unity Catalog) Link Prompt Versions to Models\x12\xd3\x01\n\x13LinkPromptsToTraces\x12/.mlflow.unitycatalog.LinkPromptsToTracesRequest\x1a\x16.google.protobuf.Empty\"s\xf2\x86\x19o\nC\n\x04POST\x12\x35/mlflow/unity-catalog/prompt-versions/links-to-traces\x1a\x04\x08\x02\x10\x00\x10\x03*&(Unity Catalog) Link Prompts to Traces\x12\xe1\x01\n\x18LinkPromptVersionsToRuns\x12\x34.mlflow.unitycatalog.LinkPromptVersionsToRunsRequest\x1a\x16.google.protobuf.Empty\"w\xf2\x86\x19s\nA\n\x04POST\x12\x33/mlflow/unity-catalog/prompt-versions/links-to-runs\x1a\x04\x08\x02\x10\x00\x10\x03*,(Unity Catalog) Link Prompt Versions to RunsB4\n\'com.databricks.api.proto.managedcatalog\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')

  _globals = globals()
  _builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
  _builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'unity_catalog_prompt_service_pb2', _globals)
  if not _descriptor._USE_C_DESCRIPTORS:
    _globals['DESCRIPTOR']._loaded_options = None
    _globals['DESCRIPTOR']._serialized_options = b'\n\'com.databricks.api.proto.managedcatalog\220\001\001\240\001\001\342?\002\020\001'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['CreatePrompt']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['CreatePrompt']._serialized_options = b'\362\206\031N\n+\n\004POST\022\035/mlflow/unity-catalog/prompts\032\004\010\002\020\000\020\003*\035(Unity Catalog) Create Prompt'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['UpdatePrompt']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['UpdatePrompt']._serialized_options = b'\362\206\031V\n3\n\005PATCH\022$/mlflow/unity-catalog/prompts/{name}\032\004\010\002\020\000\020\003*\035(Unity Catalog) Update Prompt'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['DeletePrompt']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['DeletePrompt']._serialized_options = b'\362\206\031W\n4\n\006DELETE\022$/mlflow/unity-catalog/prompts/{name}\032\004\010\002\020\000\020\003*\035(Unity Catalog) Delete Prompt'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['GetPrompt']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['GetPrompt']._serialized_options = b'\362\206\031Q\n1\n\003GET\022$/mlflow/unity-catalog/prompts/{name}\032\004\010\002\020\000\020\003*\032(Unity Catalog) Get Prompt'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['SearchPrompts']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['SearchPrompts']._serialized_options = b'\362\206\031V\n2\n\004POST\022$/mlflow/unity-catalog/prompts/search\032\004\010\002\020\000\020\003*\036(Unity Catalog) Search Prompts'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['CreatePromptVersion']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['CreatePromptVersion']._serialized_options = b'\362\206\031f\n;\n\004POST\022-/mlflow/unity-catalog/prompts/{name}/versions\032\004\010\002\020\000\020\003*%(Unity Catalog) Create Prompt Version'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['UpdatePromptVersion']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['UpdatePromptVersion']._serialized_options = b'\362\206\031q\nF\n\005PATCH\0227/mlflow/unity-catalog/prompts/{name}/versions/{version}\032\004\010\002\020\000\020\003*%(Unity Catalog) Update Prompt Version'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['DeletePromptVersion']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['DeletePromptVersion']._serialized_options = b'\362\206\031r\nG\n\006DELETE\0227/mlflow/unity-catalog/prompts/{name}/versions/{version}\032\004\010\002\020\000\020\003*%(Unity Catalog) Delete Prompt Version'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['GetPromptVersion']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['GetPromptVersion']._serialized_options = b'\362\206\031l\nD\n\003GET\0227/mlflow/unity-catalog/prompts/{name}/versions/{version}\032\004\010\002\020\000\020\003*\"(Unity Catalog) Get Prompt Version'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['SearchPromptVersions']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['SearchPromptVersions']._serialized_options = b'\362\206\031y\nB\n\004POST\0224/mlflow/unity-catalog/prompts/{name}/versions/search\032\004\010\002\020\000\020\003\030\350\007\030\354\007\030\272\027\030\001*&(Unity Catalog) Search Prompt Versions'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['SetPromptAlias']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['SetPromptAlias']._serialized_options = b'\362\206\031h\nB\n\004POST\0224/mlflow/unity-catalog/prompts/{name}/aliases/{alias}\032\004\010\002\020\000\020\003* (Unity Catalog) Set Prompt Alias'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['DeletePromptAlias']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['DeletePromptAlias']._serialized_options = b'\362\206\031m\nD\n\006DELETE\0224/mlflow/unity-catalog/prompts/{name}/aliases/{alias}\032\004\010\002\020\000\020\003*#(Unity Catalog) Delete Prompt Alias'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['GetPromptVersionByAlias']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['GetPromptVersionByAlias']._serialized_options = b'\362\206\031|\nK\n\003GET\022>/mlflow/unity-catalog/prompts/{name}/versions/by-alias/{alias}\032\004\010\002\020\000\020\003*+(Unity Catalog) Get Prompt Version By Alias'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['SetPromptTag']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['SetPromptTag']._serialized_options = b'\362\206\031[\n7\n\004POST\022)/mlflow/unity-catalog/prompts/{name}/tags\032\004\010\002\020\000\020\003*\036(Unity Catalog) Set Prompt Tag'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['DeletePromptTag']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['DeletePromptTag']._serialized_options = b'\362\206\031f\n?\n\006DELETE\022//mlflow/unity-catalog/prompts/{name}/tags/{key}\032\004\010\002\020\000\020\003*!(Unity Catalog) Delete Prompt Tag'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['SetPromptVersionTag']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['SetPromptVersionTag']._serialized_options = b'\362\206\031v\nJ\n\004POST\022</mlflow/unity-catalog/prompts/{name}/versions/{version}/tags\032\004\010\002\020\000\020\003*&(Unity Catalog) Set Prompt Version Tag'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['DeletePromptVersionTag']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['DeletePromptVersionTag']._serialized_options = b'\362\206\031\201\001\nR\n\006DELETE\022B/mlflow/unity-catalog/prompts/{name}/versions/{version}/tags/{key}\032\004\010\002\020\000\020\003*)(Unity Catalog) Delete Prompt Version Tag'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['LinkPromptVersionsToModels']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['LinkPromptVersionsToModels']._serialized_options = b'\362\206\031w\nC\n\004POST\0225/mlflow/unity-catalog/prompt-versions/links-to-models\032\004\010\002\020\000\020\003*.(Unity Catalog) Link Prompt Versions to Models'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['LinkPromptsToTraces']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['LinkPromptsToTraces']._serialized_options = b'\362\206\031o\nC\n\004POST\0225/mlflow/unity-catalog/prompt-versions/links-to-traces\032\004\010\002\020\000\020\003*&(Unity Catalog) Link Prompts to Traces'
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['LinkPromptVersionsToRuns']._loaded_options = None
    _globals['_UNITYCATALOGPROMPTSERVICE'].methods_by_name['LinkPromptVersionsToRuns']._serialized_options = b'\362\206\031s\nA\n\004POST\0223/mlflow/unity-catalog/prompt-versions/links-to-runs\032\004\010\002\020\000\020\003*,(Unity Catalog) Link Prompt Versions to Runs'
    _globals['_UNITYCATALOGPROMPTSERVICE']._serialized_start=167
    _globals['_UNITYCATALOGPROMPTSERVICE']._serialized_end=4389
  _builder.BuildServices(DESCRIPTOR, 'unity_catalog_prompt_service_pb2', _globals)
  # @@protoc_insertion_point(module_scope)

else:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: unity_catalog_prompt_service.proto
  """Generated protocol buffer code."""
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import message as _message
  from google.protobuf import reflection as _reflection
  from google.protobuf import symbol_database as _symbol_database
  from google.protobuf import service as _service
  from google.protobuf import service_reflection
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()


  from . import databricks_pb2 as databricks__pb2
  from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
  from . import unity_catalog_prompt_messages_pb2 as unity_catalog_prompt_messages_pb2
  from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"unity_catalog_prompt_service.proto\x12\x13mlflow.unitycatalog\x1a\x10\x64\x61tabricks.proto\x1a\x15scalapb/scalapb.proto\x1a#unity_catalog_prompt_messages.proto\x1a\x1bgoogle/protobuf/empty.proto2\xfe \n\x19UnityCatalogPromptService\x12\xa9\x01\n\x0c\x43reatePrompt\x12(.mlflow.unitycatalog.CreatePromptRequest\x1a\x1b.mlflow.unitycatalog.Prompt\"R\xf2\x86\x19N\n+\n\x04POST\x12\x1d/mlflow/unity-catalog/prompts\x1a\x04\x08\x02\x10\x00\x10\x03*\x1d(Unity Catalog) Create Prompt\x12\xb1\x01\n\x0cUpdatePrompt\x12(.mlflow.unitycatalog.UpdatePromptRequest\x1a\x1b.mlflow.unitycatalog.Prompt\"Z\xf2\x86\x19V\n3\n\x05PATCH\x12$/mlflow/unity-catalog/prompts/{name}\x1a\x04\x08\x02\x10\x00\x10\x03*\x1d(Unity Catalog) Update Prompt\x12\xad\x01\n\x0c\x44\x65letePrompt\x12(.mlflow.unitycatalog.DeletePromptRequest\x1a\x16.google.protobuf.Empty\"[\xf2\x86\x19W\n4\n\x06\x44\x45LETE\x12$/mlflow/unity-catalog/prompts/{name}\x1a\x04\x08\x02\x10\x00\x10\x03*\x1d(Unity Catalog) Delete Prompt\x12\xa6\x01\n\tGetPrompt\x12%.mlflow.unitycatalog.GetPromptRequest\x1a\x1b.mlflow.unitycatalog.Prompt\"U\xf2\x86\x19Q\n1\n\x03GET\x12$/mlflow/unity-catalog/prompts/{name}\x1a\x04\x08\x02\x10\x00\x10\x03*\x1a(Unity Catalog) Get Prompt\x12\xc2\x01\n\rSearchPrompts\x12).mlflow.unitycatalog.SearchPromptsRequest\x1a*.mlflow.unitycatalog.SearchPromptsResponse\"Z\xf2\x86\x19V\n2\n\x04POST\x12$/mlflow/unity-catalog/prompts/search\x1a\x04\x08\x02\x10\x00\x10\x03*\x1e(Unity Catalog) Search Prompts\x12\xd6\x01\n\x13\x43reatePromptVersion\x12/.mlflow.unitycatalog.CreatePromptVersionRequest\x1a\".mlflow.unitycatalog.PromptVersion\"j\xf2\x86\x19\x66\n;\n\x04POST\x12-/mlflow/unity-catalog/prompts/{name}/versions\x1a\x04\x08\x02\x10\x00\x10\x03*%(Unity Catalog) Create Prompt Version\x12\xe1\x01\n\x13UpdatePromptVersion\x12/.mlflow.unitycatalog.UpdatePromptVersionRequest\x1a\".mlflow.unitycatalog.PromptVersion\"u\xf2\x86\x19q\nF\n\x05PATCH\x12\x37/mlflow/unity-catalog/prompts/{name}/versions/{version}\x1a\x04\x08\x02\x10\x00\x10\x03*%(Unity Catalog) Update Prompt Version\x12\xd6\x01\n\x13\x44\x65letePromptVersion\x12/.mlflow.unitycatalog.DeletePromptVersionRequest\x1a\x16.google.protobuf.Empty\"v\xf2\x86\x19r\nG\n\x06\x44\x45LETE\x12\x37/mlflow/unity-catalog/prompts/{name}/versions/{version}\x1a\x04\x08\x02\x10\x00\x10\x03*%(Unity Catalog) Delete Prompt Version\x12\xd6\x01\n\x10GetPromptVersion\x12,.mlflow.unitycatalog.GetPromptVersionRequest\x1a\".mlflow.unitycatalog.PromptVersion\"p\xf2\x86\x19l\nD\n\x03GET\x12\x37/mlflow/unity-catalog/prompts/{name}/versions/{version}\x1a\x04\x08\x02\x10\x00\x10\x03*\"(Unity Catalog) Get Prompt Version\x12\xfa\x01\n\x14SearchPromptVersions\x12\x30.mlflow.unitycatalog.SearchPromptVersionsRequest\x1a\x31.mlflow.unitycatalog.SearchPromptVersionsResponse\"}\xf2\x86\x19y\nB\n\x04POST\x12\x34/mlflow/unity-catalog/prompts/{name}/versions/search\x1a\x04\x08\x02\x10\x00\x10\x03\x18\xe8\x07\x18\xec\x07\x18\xba\x17\x18\x01*&(Unity Catalog) Search Prompt Versions\x12\xc2\x01\n\x0eSetPromptAlias\x12*.mlflow.unitycatalog.SetPromptAliasRequest\x1a\x16.google.protobuf.Empty\"l\xf2\x86\x19h\nB\n\x04POST\x12\x34/mlflow/unity-catalog/prompts/{name}/aliases/{alias}\x1a\x04\x08\x02\x10\x00\x10\x03* (Unity Catalog) Set Prompt Alias\x12\xcd\x01\n\x11\x44\x65letePromptAlias\x12-.mlflow.unitycatalog.DeletePromptAliasRequest\x1a\x16.google.protobuf.Empty\"q\xf2\x86\x19m\nD\n\x06\x44\x45LETE\x12\x34/mlflow/unity-catalog/prompts/{name}/aliases/{alias}\x1a\x04\x08\x02\x10\x00\x10\x03*#(Unity Catalog) Delete Prompt Alias\x12\xf5\x01\n\x17GetPromptVersionByAlias\x12\x33.mlflow.unitycatalog.GetPromptVersionByAliasRequest\x1a\".mlflow.unitycatalog.PromptVersion\"\x80\x01\xf2\x86\x19|\nK\n\x03GET\x12>/mlflow/unity-catalog/prompts/{name}/versions/by-alias/{alias}\x1a\x04\x08\x02\x10\x00\x10\x03*+(Unity Catalog) Get Prompt Version By Alias\x12\xb1\x01\n\x0cSetPromptTag\x12(.mlflow.unitycatalog.SetPromptTagRequest\x1a\x16.google.protobuf.Empty\"_\xf2\x86\x19[\n7\n\x04POST\x12)/mlflow/unity-catalog/prompts/{name}/tags\x1a\x04\x08\x02\x10\x00\x10\x03*\x1e(Unity Catalog) Set Prompt Tag\x12\xc2\x01\n\x0f\x44\x65letePromptTag\x12+.mlflow.unitycatalog.DeletePromptTagRequest\x1a\x16.google.protobuf.Empty\"j\xf2\x86\x19\x66\n?\n\x06\x44\x45LETE\x12//mlflow/unity-catalog/prompts/{name}/tags/{key}\x1a\x04\x08\x02\x10\x00\x10\x03*!(Unity Catalog) Delete Prompt Tag\x12\xda\x01\n\x13SetPromptVersionTag\x12/.mlflow.unitycatalog.SetPromptVersionTagRequest\x1a\x16.google.protobuf.Empty\"z\xf2\x86\x19v\nJ\n\x04POST\x12</mlflow/unity-catalog/prompts/{name}/versions/{version}/tags\x1a\x04\x08\x02\x10\x00\x10\x03*&(Unity Catalog) Set Prompt Version Tag\x12\xed\x01\n\x16\x44\x65letePromptVersionTag\x12\x32.mlflow.unitycatalog.DeletePromptVersionTagRequest\x1a\x16.google.protobuf.Empty\"\x86\x01\xf2\x86\x19\x81\x01\nR\n\x06\x44\x45LETE\x12\x42/mlflow/unity-catalog/prompts/{name}/versions/{version}/tags/{key}\x1a\x04\x08\x02\x10\x00\x10\x03*)(Unity Catalog) Delete Prompt Version Tag\x12\xe9\x01\n\x1aLinkPromptVersionsToModels\x12\x36.mlflow.unitycatalog.LinkPromptVersionsToModelsRequest\x1a\x16.google.protobuf.Empty\"{\xf2\x86\x19w\nC\n\x04POST\x12\x35/mlflow/unity-catalog/prompt-versions/links-to-models\x1a\x04\x08\x02\x10\x00\x10\x03*.(Unity Catalog) Link Prompt Versions to Models\x12\xd3\x01\n\x13LinkPromptsToTraces\x12/.mlflow.unitycatalog.LinkPromptsToTracesRequest\x1a\x16.google.protobuf.Empty\"s\xf2\x86\x19o\nC\n\x04POST\x12\x35/mlflow/unity-catalog/prompt-versions/links-to-traces\x1a\x04\x08\x02\x10\x00\x10\x03*&(Unity Catalog) Link Prompts to Traces\x12\xe1\x01\n\x18LinkPromptVersionsToRuns\x12\x34.mlflow.unitycatalog.LinkPromptVersionsToRunsRequest\x1a\x16.google.protobuf.Empty\"w\xf2\x86\x19s\nA\n\x04POST\x12\x33/mlflow/unity-catalog/prompt-versions/links-to-runs\x1a\x04\x08\x02\x10\x00\x10\x03*,(Unity Catalog) Link Prompt Versions to RunsB4\n\'com.databricks.api.proto.managedcatalog\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')



  _UNITYCATALOGPROMPTSERVICE = DESCRIPTOR.services_by_name['UnityCatalogPromptService']
  if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b'\n\'com.databricks.api.proto.managedcatalog\220\001\001\240\001\001\342?\002\020\001'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['CreatePrompt']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['CreatePrompt']._serialized_options = b'\362\206\031N\n+\n\004POST\022\035/mlflow/unity-catalog/prompts\032\004\010\002\020\000\020\003*\035(Unity Catalog) Create Prompt'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['UpdatePrompt']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['UpdatePrompt']._serialized_options = b'\362\206\031V\n3\n\005PATCH\022$/mlflow/unity-catalog/prompts/{name}\032\004\010\002\020\000\020\003*\035(Unity Catalog) Update Prompt'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['DeletePrompt']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['DeletePrompt']._serialized_options = b'\362\206\031W\n4\n\006DELETE\022$/mlflow/unity-catalog/prompts/{name}\032\004\010\002\020\000\020\003*\035(Unity Catalog) Delete Prompt'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['GetPrompt']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['GetPrompt']._serialized_options = b'\362\206\031Q\n1\n\003GET\022$/mlflow/unity-catalog/prompts/{name}\032\004\010\002\020\000\020\003*\032(Unity Catalog) Get Prompt'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['SearchPrompts']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['SearchPrompts']._serialized_options = b'\362\206\031V\n2\n\004POST\022$/mlflow/unity-catalog/prompts/search\032\004\010\002\020\000\020\003*\036(Unity Catalog) Search Prompts'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['CreatePromptVersion']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['CreatePromptVersion']._serialized_options = b'\362\206\031f\n;\n\004POST\022-/mlflow/unity-catalog/prompts/{name}/versions\032\004\010\002\020\000\020\003*%(Unity Catalog) Create Prompt Version'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['UpdatePromptVersion']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['UpdatePromptVersion']._serialized_options = b'\362\206\031q\nF\n\005PATCH\0227/mlflow/unity-catalog/prompts/{name}/versions/{version}\032\004\010\002\020\000\020\003*%(Unity Catalog) Update Prompt Version'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['DeletePromptVersion']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['DeletePromptVersion']._serialized_options = b'\362\206\031r\nG\n\006DELETE\0227/mlflow/unity-catalog/prompts/{name}/versions/{version}\032\004\010\002\020\000\020\003*%(Unity Catalog) Delete Prompt Version'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['GetPromptVersion']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['GetPromptVersion']._serialized_options = b'\362\206\031l\nD\n\003GET\0227/mlflow/unity-catalog/prompts/{name}/versions/{version}\032\004\010\002\020\000\020\003*\"(Unity Catalog) Get Prompt Version'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['SearchPromptVersions']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['SearchPromptVersions']._serialized_options = b'\362\206\031y\nB\n\004POST\0224/mlflow/unity-catalog/prompts/{name}/versions/search\032\004\010\002\020\000\020\003\030\350\007\030\354\007\030\272\027\030\001*&(Unity Catalog) Search Prompt Versions'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['SetPromptAlias']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['SetPromptAlias']._serialized_options = b'\362\206\031h\nB\n\004POST\0224/mlflow/unity-catalog/prompts/{name}/aliases/{alias}\032\004\010\002\020\000\020\003* (Unity Catalog) Set Prompt Alias'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['DeletePromptAlias']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['DeletePromptAlias']._serialized_options = b'\362\206\031m\nD\n\006DELETE\0224/mlflow/unity-catalog/prompts/{name}/aliases/{alias}\032\004\010\002\020\000\020\003*#(Unity Catalog) Delete Prompt Alias'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['GetPromptVersionByAlias']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['GetPromptVersionByAlias']._serialized_options = b'\362\206\031|\nK\n\003GET\022>/mlflow/unity-catalog/prompts/{name}/versions/by-alias/{alias}\032\004\010\002\020\000\020\003*+(Unity Catalog) Get Prompt Version By Alias'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['SetPromptTag']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['SetPromptTag']._serialized_options = b'\362\206\031[\n7\n\004POST\022)/mlflow/unity-catalog/prompts/{name}/tags\032\004\010\002\020\000\020\003*\036(Unity Catalog) Set Prompt Tag'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['DeletePromptTag']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['DeletePromptTag']._serialized_options = b'\362\206\031f\n?\n\006DELETE\022//mlflow/unity-catalog/prompts/{name}/tags/{key}\032\004\010\002\020\000\020\003*!(Unity Catalog) Delete Prompt Tag'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['SetPromptVersionTag']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['SetPromptVersionTag']._serialized_options = b'\362\206\031v\nJ\n\004POST\022</mlflow/unity-catalog/prompts/{name}/versions/{version}/tags\032\004\010\002\020\000\020\003*&(Unity Catalog) Set Prompt Version Tag'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['DeletePromptVersionTag']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['DeletePromptVersionTag']._serialized_options = b'\362\206\031\201\001\nR\n\006DELETE\022B/mlflow/unity-catalog/prompts/{name}/versions/{version}/tags/{key}\032\004\010\002\020\000\020\003*)(Unity Catalog) Delete Prompt Version Tag'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['LinkPromptVersionsToModels']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['LinkPromptVersionsToModels']._serialized_options = b'\362\206\031w\nC\n\004POST\0225/mlflow/unity-catalog/prompt-versions/links-to-models\032\004\010\002\020\000\020\003*.(Unity Catalog) Link Prompt Versions to Models'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['LinkPromptsToTraces']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['LinkPromptsToTraces']._serialized_options = b'\362\206\031o\nC\n\004POST\0225/mlflow/unity-catalog/prompt-versions/links-to-traces\032\004\010\002\020\000\020\003*&(Unity Catalog) Link Prompts to Traces'
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['LinkPromptVersionsToRuns']._options = None
    _UNITYCATALOGPROMPTSERVICE.methods_by_name['LinkPromptVersionsToRuns']._serialized_options = b'\362\206\031s\nA\n\004POST\0223/mlflow/unity-catalog/prompt-versions/links-to-runs\032\004\010\002\020\000\020\003*,(Unity Catalog) Link Prompt Versions to Runs'
    _UNITYCATALOGPROMPTSERVICE._serialized_start=167
    _UNITYCATALOGPROMPTSERVICE._serialized_end=4389
  UnityCatalogPromptService = service_reflection.GeneratedServiceType('UnityCatalogPromptService', (_service.Service,), dict(
    DESCRIPTOR = _UNITYCATALOGPROMPTSERVICE,
    __module__ = 'unity_catalog_prompt_service_pb2'
    ))

  UnityCatalogPromptService_Stub = service_reflection.GeneratedServiceStubType('UnityCatalogPromptService_Stub', (UnityCatalogPromptService,), dict(
    DESCRIPTOR = _UNITYCATALOGPROMPTSERVICE,
    __module__ = 'unity_catalog_prompt_service_pb2'
    ))


  # @@protoc_insertion_point(module_scope)




"""
Plugin Manager for Genesis-Flow

Central management system for all plugins including loading, enabling,
disabling, and lifecycle management.
"""

import logging
import threading
from typing import Dict, List, Optional, Any, Type
from contextlib import contextmanager

from mlflow.plugins.base import BasePlugin, PluginState, PluginType
from mlflow.plugins.discovery import PluginDiscovery, discover_plugins
from mlflow.plugins.registry import PluginRegistry
from mlflow.version import VERSION as GENESIS_FLOW_VERSION

logger = logging.getLogger(__name__)

class PluginManager:
    """
    Central plugin management system for Genesis-Flow.
    
    Handles plugin discovery, loading, enabling/disabling, and lifecycle management.
    Thread-safe operations with proper error handling and state management.
    """
    
    def __init__(self):
        self._registry = PluginRegistry()
        self._discovery = PluginDiscovery()
        self._instances: Dict[str, BasePlugin] = {}
        self._lock = threading.RLock()
        self._initialized = False
        
    def initialize(self, auto_discover: bool = True, auto_enable_builtin: bool = True):
        """
        Initialize the plugin manager.
        
        Args:
            auto_discover: Whether to automatically discover all plugins
            auto_enable_builtin: Whether to automatically enable built-in plugins
        """
        with self._lock:
            if self._initialized:
                return
            
            try:
                logger.info("Initializing Genesis-Flow plugin manager")
                
                if auto_discover:
                    self._discover_plugins()
                
                if auto_enable_builtin:
                    self._enable_builtin_plugins()
                
                self._initialized = True
                logger.info(f"Plugin manager initialized with {len(self._instances)} plugins")
                
            except Exception as e:
                logger.error(f"Failed to initialize plugin manager: {e}")
                raise
    
    def _discover_plugins(self):
        """Discover all available plugins."""
        discovered = self._discovery.discover_all_plugins()
        
        for name, plugin_class in discovered.items():
            try:
                self._registry.register_plugin_class(name, plugin_class)
                logger.debug(f"Registered plugin class: {name}")
            except Exception as e:
                logger.error(f"Failed to register plugin {name}: {e}")
    
    def _enable_builtin_plugins(self):
        """Enable all built-in plugins that are compatible."""
        builtin_plugins = self._discovery.discover_builtin_plugins()
        
        for name, plugin_class in builtin_plugins.items():
            try:
                if self._is_plugin_compatible(name):
                    self.enable_plugin(name)
                    logger.debug(f"Auto-enabled built-in plugin: {name}")
                else:
                    logger.warning(f"Built-in plugin {name} is not compatible")
            except Exception as e:
                logger.error(f"Failed to auto-enable built-in plugin {name}: {e}")
    
    def list_plugins(self, plugin_type: Optional[PluginType] = None, 
                    state: Optional[PluginState] = None) -> List[Dict[str, Any]]:
        """
        List all plugins with optional filtering.
        
        Args:
            plugin_type: Filter by plugin type
            state: Filter by plugin state
            
        Returns:
            List of plugin information dictionaries
        """
        with self._lock:
            plugins = []
            
            # Include registered but not instantiated plugins
            for name in self._registry.list_plugins():
                plugin_info = self._get_plugin_info(name)
                if plugin_info:
                    # Apply filters
                    if plugin_type and plugin_info.get('type') != plugin_type.value:
                        continue
                    if state and plugin_info.get('state') != state.value:
                        continue
                    
                    plugins.append(plugin_info)
            
            return plugins
    
    def get_plugin(self, name: str) -> Optional[BasePlugin]:
        """
        Get a plugin instance by name.
        
        Args:
            name: Plugin name
            
        Returns:
            Plugin instance or None if not found/enabled
        """
        with self._lock:
            return self._instances.get(name)
    
    def enable_plugin(self, name: str) -> bool:
        """
        Enable a plugin by name.
        
        Args:
            name: Plugin name
            
        Returns:
            True if successfully enabled, False otherwise
        """
        with self._lock:
            try:
                # Check if already enabled
                if name in self._instances:
                    instance = self._instances[name]
                    if instance.state == PluginState.ENABLED:
                        logger.debug(f"Plugin {name} already enabled")
                        return True
                
                # Get plugin class from registry
                plugin_class = self._registry.get_plugin_class(name)
                if not plugin_class:
                    logger.error(f"Plugin {name} not found in registry")
                    return False
                
                # Create instance if not exists
                if name not in self._instances:
                    try:
                        # Get metadata for the plugin
                        metadata = self._get_plugin_metadata(name, plugin_class)
                        instance = plugin_class(metadata)
                        self._instances[name] = instance
                    except Exception as e:
                        logger.error(f"Failed to create instance of plugin {name}: {e}")
                        return False
                
                instance = self._instances[name]
                
                # Check compatibility
                if not instance.is_compatible(GENESIS_FLOW_VERSION):
                    logger.error(f"Plugin {name} is not compatible with Genesis-Flow {GENESIS_FLOW_VERSION}")
                    return False
                
                # Check dependencies
                if not instance.check_dependencies():
                    logger.error(f"Plugin {name} dependencies not satisfied")
                    return False
                
                # Load if not loaded
                if instance.state == PluginState.DISCOVERED:
                    if not instance.load():
                        logger.error(f"Failed to load plugin {name}")
                        return False
                
                # Enable the plugin
                if instance.enable():
                    logger.info(f"Plugin {name} enabled successfully")
                    return True
                else:
                    logger.error(f"Failed to enable plugin {name}")
                    return False
                
            except Exception as e:
                logger.error(f"Error enabling plugin {name}: {e}")
                return False
    
    def disable_plugin(self, name: str) -> bool:
        """
        Disable a plugin by name.
        
        Args:
            name: Plugin name
            
        Returns:
            True if successfully disabled, False otherwise
        """
        with self._lock:
            try:
                if name not in self._instances:
                    logger.warning(f"Plugin {name} not found or not enabled")
                    return True  # Already disabled
                
                instance = self._instances[name]
                
                if instance.disable():
                    logger.info(f"Plugin {name} disabled successfully")
                    return True
                else:
                    logger.error(f"Failed to disable plugin {name}")
                    return False
                
            except Exception as e:
                logger.error(f"Error disabling plugin {name}: {e}")
                return False
    
    def unload_plugin(self, name: str) -> bool:
        """
        Unload a plugin by name.
        
        Args:
            name: Plugin name
            
        Returns:
            True if successfully unloaded, False otherwise
        """
        with self._lock:
            try:
                if name not in self._instances:
                    logger.warning(f"Plugin {name} not found")
                    return True  # Already unloaded
                
                instance = self._instances[name]
                
                if instance.unload():
                    del self._instances[name]
                    logger.info(f"Plugin {name} unloaded successfully")
                    return True
                else:
                    logger.error(f"Failed to unload plugin {name}")
                    return False
                
            except Exception as e:
                logger.error(f"Error unloading plugin {name}: {e}")
                return False
    
    def reload_plugin(self, name: str) -> bool:
        """
        Reload a plugin by name.
        
        Args:
            name: Plugin name
            
        Returns:
            True if successfully reloaded, False otherwise
        """
        with self._lock:
            try:
                # Unload first
                if name in self._instances:
                    if not self.unload_plugin(name):
                        return False
                
                # Re-enable
                return self.enable_plugin(name)
                
            except Exception as e:
                logger.error(f"Error reloading plugin {name}: {e}")
                return False
    
    def is_plugin_enabled(self, name: str) -> bool:
        """
        Check if a plugin is enabled.
        
        Args:
            name: Plugin name
            
        Returns:
            True if enabled, False otherwise
        """
        with self._lock:
            if name in self._instances:
                return self._instances[name].state == PluginState.ENABLED
            return False
    
    def is_plugin_available(self, name: str) -> bool:
        """
        Check if a plugin is available (registered).
        
        Args:
            name: Plugin name
            
        Returns:
            True if available, False otherwise
        """
        return self._registry.is_plugin_registered(name)
    
    def _is_plugin_compatible(self, name: str) -> bool:
        """Check if a plugin is compatible with current environment."""
        try:
            plugin_class = self._registry.get_plugin_class(name)
            if not plugin_class:
                return False
            
            # Create temporary instance to check compatibility
            metadata = self._get_plugin_metadata(name, plugin_class)
            temp_instance = plugin_class(metadata)
            
            return (temp_instance.is_compatible(GENESIS_FLOW_VERSION) and 
                   temp_instance.check_dependencies())
            
        except Exception as e:
            logger.error(f"Error checking compatibility for plugin {name}: {e}")
            return False
    
    def _get_plugin_metadata(self, name: str, plugin_class: Type[BasePlugin]):
        """Get or create metadata for a plugin."""
        try:
            # Try to get metadata from plugin class
            if hasattr(plugin_class, 'get_default_metadata'):
                return plugin_class.get_default_metadata()
            
            # Create basic metadata
            from mlflow.plugins.base import PluginMetadata, PluginType
            return PluginMetadata(
                name=name,
                version="1.0.0",
                description=f"Plugin {name}",
                author="Genesis-Flow",
                plugin_type=PluginType.CUSTOM,
                dependencies=[],
                optional_dependencies=[],
                min_genesis_flow_version="1.0.0",
            )
            
        except Exception as e:
            logger.error(f"Error getting metadata for plugin {name}: {e}")
            raise
    
    def _get_plugin_info(self, name: str) -> Optional[Dict[str, Any]]:
        """Get information about a plugin."""
        try:
            if name in self._instances:
                return self._instances[name].get_info()
            
            # Get info from plugin class
            plugin_class = self._registry.get_plugin_class(name)
            if plugin_class:
                try:
                    metadata = self._get_plugin_metadata(name, plugin_class)
                    return {
                        "name": metadata.name,
                        "version": metadata.version,
                        "description": metadata.description,
                        "author": metadata.author,
                        "type": metadata.plugin_type.value,
                        "state": PluginState.DISCOVERED.value,
                        "dependencies": metadata.dependencies,
                        "optional_dependencies": metadata.optional_dependencies,
                        "tags": metadata.tags,
                    }
                except Exception:
                    return {
                        "name": name,
                        "class": plugin_class.__name__,
                        "state": PluginState.DISCOVERED.value,
                    }
            
            return None
            
        except Exception as e:
            logger.error(f"Error getting info for plugin {name}: {e}")
            return None
    
    @contextmanager
    def plugin_context(self, name: str):
        """
        Context manager for temporarily enabling a plugin.
        
        Args:
            name: Plugin name
        """
        was_enabled = self.is_plugin_enabled(name)
        
        try:
            if not was_enabled:
                if not self.enable_plugin(name):
                    raise RuntimeError(f"Failed to enable plugin {name}")
            
            yield self.get_plugin(name)
            
        finally:
            if not was_enabled and self.is_plugin_enabled(name):
                self.disable_plugin(name)
    
    def get_plugins_by_type(self, plugin_type: PluginType) -> List[BasePlugin]:
        """
        Get all enabled plugins of a specific type.
        
        Args:
            plugin_type: Type of plugins to get
            
        Returns:
            List of enabled plugin instances
        """
        with self._lock:
            plugins = []
            for instance in self._instances.values():
                if (instance.state == PluginState.ENABLED and 
                    instance.metadata.plugin_type == plugin_type):
                    plugins.append(instance)
            return plugins
    
    def get_stats(self) -> Dict[str, Any]:
        """
        Get plugin manager statistics.
        
        Returns:
            Statistics dictionary
        """
        with self._lock:
            stats = {
                "total_registered": len(self._registry.list_plugins()),
                "total_instances": len(self._instances),
                "enabled": sum(1 for p in self._instances.values() 
                             if p.state == PluginState.ENABLED),
                "disabled": sum(1 for p in self._instances.values() 
                              if p.state == PluginState.DISABLED),
                "failed": sum(1 for p in self._instances.values() 
                            if p.state == PluginState.FAILED),
                "by_type": {},
            }
            
            # Count by type
            for instance in self._instances.values():
                plugin_type = instance.metadata.plugin_type.value
                if plugin_type not in stats["by_type"]:
                    stats["by_type"][plugin_type] = 0
                stats["by_type"][plugin_type] += 1
            
            return stats
    
    def shutdown(self):
        """Shutdown the plugin manager and unload all plugins."""
        with self._lock:
            logger.info("Shutting down plugin manager")
            
            # Unload all plugins
            for name in list(self._instances.keys()):
                try:
                    self.unload_plugin(name)
                except Exception as e:
                    logger.error(f"Error unloading plugin {name} during shutdown: {e}")
            
            self._instances.clear()
            self._initialized = False
            logger.info("Plugin manager shutdown complete")


from typing import Any, Generator, Optional

import pydantic

from mlflow.exceptions import MlflowException
from mlflow.models.utils import _convert_llm_ndarray_to_list
from mlflow.protos.databricks_pb2 import INTERNAL_ERROR
from mlflow.pyfunc.model import _load_context_model_and_signature
from mlflow.types.type_hints import model_validate
from mlflow.utils.annotations import experimental
from mlflow.utils.pydantic_utils import IS_PYDANTIC_V2_OR_NEWER

if not IS_PYDANTIC_V2_OR_NEWER:
    raise ImportError(
        "ResponsesAgent and its pydantic classes are not supported in pydantic v1. "
        "Please upgrade to pydantic v2 or newer to use ResponsesAgent.",
    )

from mlflow.types.responses import (
    ResponsesAgentRequest,
    ResponsesAgentResponse,
    ResponsesAgentStreamEvent,
)


def _load_pyfunc(model_path: str, model_config: Optional[dict[str, Any]] = None):
    _, responses_agent, _ = _load_context_model_and_signature(model_path, model_config)
    return _ResponsesAgentPyfuncWrapper(responses_agent)


@experimental(version="3.0.0")
class _ResponsesAgentPyfuncWrapper:
    """
    Wrapper class that converts dict inputs to pydantic objects accepted by
    :class:`~ResponsesAgent`.
    """

    def __init__(self, responses_agent):
        self.responses_agent = responses_agent

    def get_raw_model(self):
        """
        Returns the underlying model.
        """
        return self.responses_agent

    def _convert_input(self, model_input) -> ResponsesAgentRequest:
        import pandas

        if isinstance(model_input, pandas.DataFrame):
            model_input = {
                k: _convert_llm_ndarray_to_list(v[0])
                for k, v in model_input.to_dict(orient="list").items()
            }
        elif not isinstance(model_input, dict):
            raise MlflowException(
                "Unsupported model input type. Expected a dict or pandas.DataFrame, but got "
                f"{type(model_input)} instead.",
                error_code=INTERNAL_ERROR,
            )
        return ResponsesAgentRequest(**model_input)

    def _response_to_dict(self, response, pydantic_class) -> dict[str, Any]:
        if isinstance(response, pydantic_class):
            return response.model_dump_compat(exclude_none=True)
        try:
            model_validate(pydantic_class, response)
        except pydantic.ValidationError as e:
            raise MlflowException(
                message=(
                    f"Model returned an invalid response. Expected a {pydantic_class.__name__} "
                    f"object or dictionary with the same schema. Pydantic validation error: {e}"
                ),
                error_code=INTERNAL_ERROR,
            ) from e
        return response

    def predict(self, model_input: dict[str, Any], params=None) -> dict[str, Any]:
        """
        Args:
            model_input: A dict with the
                :py:class:`ResponsesRequest <mlflow.types.responses.ResponsesRequest>` schema.
            params: Unused in this function, but required in the signature because
                `load_model_and_predict` in `utils/_capture_modules.py` expects a params field

        Returns:
            A dict with the
            (:py:class:`ResponsesResponse <mlflow.types.responses.ResponsesResponse>`)
            schema.
        """
        request = self._convert_input(model_input)
        response = self.responses_agent.predict(request)
        return self._response_to_dict(response, ResponsesAgentResponse)

    def predict_stream(
        self, model_input: dict[str, Any], params=None
    ) -> Generator[dict[str, Any], None, None]:
        """
        Args:
            model_input: A dict with the
                :py:class:`ResponsesRequest <mlflow.types.responses.ResponsesRequest>` schema.
            params: Unused in this function, but required in the signature because
                `load_model_and_predict` in `utils/_capture_modules.py` expects a params field

        Returns:
            A generator over dicts with the
                (:py:class:`ResponsesStreamEvent <mlflow.types.responses.ResponsesStreamEvent>`)
                schema.
        """
        request = self._convert_input(model_input)
        for response in self.responses_agent.predict_stream(request):
            yield self._response_to_dict(response, ResponsesAgentStreamEvent)



import os
import posixpath
import urllib.parse
from contextlib import contextmanager
from typing import Optional

try:
    from pyarrow.fs import FileSelector, FileType, HadoopFileSystem
except ImportError:
    pass

from mlflow.entities import FileInfo
from mlflow.environment_variables import (
    MLFLOW_KERBEROS_TICKET_CACHE,
    MLFLOW_KERBEROS_USER,
    MLFLOW_PYARROW_EXTRA_CONF,
)
from mlflow.store.artifact.artifact_repo import ArtifactRepository
from mlflow.utils.file_utils import relative_path_to_artifact_path


class HdfsArtifactRepository(ArtifactRepository):
    """
    Stores artifacts on HDFS.

    This repository is used with URIs of the form ``hdfs:/<path>``. The repository can only be used
    together with the RestStore.
    """

    def __init__(self, artifact_uri: str, tracking_uri: Optional[str] = None) -> None:
        super().__init__(artifact_uri, tracking_uri)
        self.scheme, self.host, self.port, self.path = _resolve_connection_params(artifact_uri)

    def log_artifact(self, local_file, artifact_path=None):
        """
        Log artifact in hdfs.

        Args:
            local_file: Source file path.
            artifact_path: When specified will attempt to write under artifact_uri/artifact_path.
        """
        hdfs_base_path = _resolve_base_path(self.path, artifact_path)

        with hdfs_system(scheme=self.scheme, host=self.host, port=self.port) as hdfs:
            _, file_name = os.path.split(local_file)
            destination_path = posixpath.join(hdfs_base_path, file_name)
            with open(local_file, "rb") as source:
                with hdfs.open_output_stream(destination_path) as destination:
                    destination.write(source.read())

    def log_artifacts(self, local_dir, artifact_path=None):
        """
        Log artifacts in hdfs.
        Missing remote sub-directories will be created if needed.

        Args:
            local_dir: Source dir path.
            artifact_path: When specified will attempt to write under artifact_uri/artifact_path.
        """
        hdfs_base_path = _resolve_base_path(self.path, artifact_path)

        with hdfs_system(scheme=self.scheme, host=self.host, port=self.port) as hdfs:
            if not hdfs.get_file_info(hdfs_base_path).type == FileType.Directory:
                hdfs.create_dir(hdfs_base_path, recursive=True)

            for subdir_path, _, files in os.walk(local_dir):
                relative_path = _relative_path_local(local_dir, subdir_path)

                hdfs_subdir_path = (
                    posixpath.join(hdfs_base_path, relative_path)
                    if relative_path
                    else hdfs_base_path
                )

                if not hdfs.get_file_info(hdfs_subdir_path).type == FileType.Directory:
                    hdfs.create_dir(hdfs_subdir_path, recursive=True)

                for each_file in files:
                    source_path = os.path.join(subdir_path, each_file)
                    destination_path = posixpath.join(hdfs_subdir_path, each_file)
                    with open(source_path, "rb") as source:
                        with hdfs.open_output_stream(destination_path) as destination:
                            destination.write(source.read())

    def list_artifacts(self, path=None):
        """
        Lists files and directories under artifacts directory for the current run_id.
        (self.path contains the base path - hdfs:/some/path/run_id/artifacts)

        Args:
            path: Relative source path. Possible subdirectory existing under
                hdfs:/some/path/run_id/artifacts

        Returns:
            List of FileInfos under given path
        """
        hdfs_base_path = _resolve_base_path(self.path, path)

        with hdfs_system(scheme=self.scheme, host=self.host, port=self.port) as hdfs:
            paths = []
            base_info = hdfs.get_file_info(hdfs_base_path)
            if base_info.type == FileType.Directory:
                selector = FileSelector(hdfs_base_path)
            elif base_info.type == FileType.File:
                selector = [hdfs_base_path]
            else:
                return []

            for file_detail in hdfs.get_file_info(selector):
                file_name = file_detail.path

                # file_name is hdfs_base_path and not a child of that path
                if file_name == hdfs_base_path:
                    continue

                # Strip off anything that comes before the artifact root e.g. hdfs://name
                offset = file_name.index(self.path)
                rel_path = _relative_path_remote(self.path, file_name[offset:])
                is_dir = file_detail.type == FileType.Directory
                size = file_detail.size
                paths.append(FileInfo(rel_path, is_dir=is_dir, file_size=size))
            return sorted(paths, key=lambda f: paths)

    def _is_directory(self, artifact_path):
        hdfs_base_path = _resolve_base_path(self.path, artifact_path)
        with hdfs_system(scheme=self.scheme, host=self.host, port=self.port) as hdfs:
            return hdfs.get_file_info(hdfs_base_path).type == FileType.Directory

    def _download_file(self, remote_file_path, local_path):
        hdfs_base_path = _resolve_base_path(self.path, remote_file_path)
        with hdfs_system(scheme=self.scheme, host=self.host, port=self.port) as hdfs:
            with hdfs.open_input_stream(hdfs_base_path) as source:
                with open(local_path, "wb") as destination:
                    destination.write(source.read())

    def delete_artifacts(self, artifact_path=None):
        path = posixpath.join(self.path, artifact_path) if artifact_path else self.path
        with hdfs_system(scheme=self.scheme, host=self.host, port=self.port) as hdfs:
            file_info = hdfs.get_file_info(path)
            if file_info.type == FileType.File:
                hdfs.delete_file(path)
            elif file_info.type == FileType.Directory:
                hdfs.delete_dir_contents(path)


@contextmanager
def hdfs_system(scheme, host, port):
    """
    hdfs system context - Attempt to establish the connection to hdfs
    and yields HadoopFileSystem

    Args:
        scheme: scheme or use hdfs:// as default
        host: hostname or when relaying on the core-site.xml config use 'default'
        port: port or when relaying on the core-site.xml config use 0
    """
    kerb_ticket = MLFLOW_KERBEROS_TICKET_CACHE.get()
    kerberos_user = MLFLOW_KERBEROS_USER.get()
    extra_conf = _parse_extra_conf(MLFLOW_PYARROW_EXTRA_CONF.get())

    host = scheme + "://" + host if host else "default"

    yield HadoopFileSystem(
        host=host,
        port=port or 0,
        user=kerberos_user,
        kerb_ticket=kerb_ticket,
        extra_conf=extra_conf,
    )


def _resolve_connection_params(artifact_uri):
    parsed = urllib.parse.urlparse(artifact_uri)

    return parsed.scheme, parsed.hostname, parsed.port, parsed.path


def _resolve_base_path(path, artifact_path):
    if path == artifact_path:
        return path
    if artifact_path:
        return posixpath.join(path, artifact_path)
    return path


def _relative_path(base_dir, subdir_path, path_module):
    relative_path = path_module.relpath(subdir_path, base_dir)
    return relative_path if relative_path != "." else None


def _relative_path_local(base_dir, subdir_path):
    rel_path = _relative_path(base_dir, subdir_path, os.path)
    return relative_path_to_artifact_path(rel_path) if rel_path is not None else None


def _relative_path_remote(base_dir, subdir_path):
    return _relative_path(base_dir, subdir_path, posixpath)


def _parse_extra_conf(extra_conf):
    if extra_conf:

        def as_pair(config):
            key, val = config.split("=")
            return key, val

        list_of_key_val = [as_pair(conf) for conf in extra_conf.split(",")]
        return dict(list_of_key_val)
    return None



from typing import TYPE_CHECKING, Any, Optional, Union

from mlflow.data import Dataset
from mlflow.data.digest_utils import compute_pandas_digest
from mlflow.data.evaluation_dataset import EvaluationDataset as LegacyEvaluationDataset
from mlflow.data.pyfunc_dataset_mixin import PyFuncConvertibleDatasetMixin
from mlflow.data.spark_dataset_source import SparkDatasetSource
from mlflow.entities import Dataset as DatasetEntity

if TYPE_CHECKING:
    import pandas as pd
    import pyspark
    from databricks.agents.datasets import Dataset as ManagedDataset


class EvaluationDataset(Dataset, PyFuncConvertibleDatasetMixin):
    """
    A dataset for storing evaluation records (inputs and expectations).

    Currently, this class is only supported for Databricks managed datasets.
    To use this class, you must have the `databricks-agents` package installed.
    """

    def __init__(self, dataset: "ManagedDataset"):
        self._dataset = dataset
        self._df = None
        self._digest = None

    @property
    def dataset_id(self) -> str:
        """The unique identifier of the dataset."""
        return self._dataset.dataset_id

    @property
    def digest(self) -> Optional[str]:
        """String digest (hash) of the dataset provided by the caller that uniquely identifies"""
        # NB: The managed Dataset entity in Agent SDK doesn't propagate the digest
        # information. So we compute the digest of the dataframe view.
        if self._digest is None:
            self._digest = self._dataset.digest or compute_pandas_digest(self.to_df())
        return self._digest

    @property
    def name(self) -> Optional[str]:
        """The UC table name of the dataset."""
        return self._dataset.name

    @property
    def schema(self) -> Optional[str]:
        """The schema of the dataset."""
        return self._dataset.schema

    @property
    def profile(self) -> Optional[str]:
        """The profile of the dataset, summary statistics."""
        return self._dataset.profile

    @property
    def source(self) -> Optional[str]:
        """Source information for the dataset."""
        # NB: The managed Dataset entity in Agent SDK doesn't propagate the source
        # information. So we use the table name as the fallback source.
        return self._dataset.source or SparkDatasetSource(table_name=self.name)

    @property
    def source_type(self) -> Optional[str]:
        """The type of the dataset source, e.g. "databricks-uc-table", "DBFS", "S3", ..."""
        return self._dataset.source_type

    @property
    def create_time(self) -> Optional[str]:
        """The time the dataset was created."""
        return self._dataset.create_time

    @property
    def created_by(self) -> Optional[str]:
        """The user who created the dataset."""
        return self._dataset.created_by

    @property
    def last_update_time(self) -> Optional[str]:
        """The time the dataset was last updated."""
        return self._dataset.last_update_time

    @property
    def last_updated_by(self) -> Optional[str]:
        """The user who last updated the dataset."""
        return self._dataset.last_updated_by

    def set_profile(self, profile: str) -> "EvaluationDataset":
        """Set the profile of the dataset."""
        dataset = self._dataset.set_profile(profile)
        return EvaluationDataset(dataset)

    def merge_records(
        self,
        records: Union[list[dict[str, Any]], "pd.DataFrame", "pyspark.sql.DataFrame"],
    ) -> "EvaluationDataset":
        """Merge records into the dataset."""
        dataset = self._dataset.merge_records(records)
        return EvaluationDataset(dataset)

    def to_df(self) -> "pd.DataFrame":
        """Convert the dataset to a pandas DataFrame."""
        # Cache the dataframe view to avoid re-fetching the records many times
        if self._df is None:
            self._df = self._dataset.to_df()
        return self._df

    def to_evaluation_dataset(self, path=None, feature_names=None) -> LegacyEvaluationDataset:
        """
        Converts the dataset to the legacy EvaluationDataset for model evaluation. Required
        for use with mlflow.evaluate().
        """
        return LegacyEvaluationDataset(
            data=self.to_df(),
            path=path,
            feature_names=feature_names,
            name=self.name,
            digest=self.digest,
        )

    def _to_mlflow_entity(self) -> DatasetEntity:
        return DatasetEntity(
            name=self.name,
            digest=self.digest,
            source_type=self.source_type,
            source=self.source.to_json(),
            schema=self.schema,
            profile=self.profile,
        )



import logging

from mlflow.exceptions import MlflowException

_logger = logging.getLogger(__name__)


def apply_span_processors(span):
    """Apply configured span processors sequentially to the span."""
    from mlflow.tracing.config import get_config

    config = get_config()
    if not config.span_processors:
        return

    non_null_return_processors = []
    for processor in config.span_processors:
        try:
            result = processor(span)
            if result is not None:
                non_null_return_processors.append(processor.__name__)
        except Exception as e:
            _logger.warning(
                f"Span processor {processor.__name__} failed: {e}",
                exc_info=_logger.isEnabledFor(logging.DEBUG),
            )

    if non_null_return_processors:
        _logger.warning(
            f"Span processors {non_null_return_processors} returned a non-null value, "
            "but it will be ignored. Span processors should not return a value."
        )


def validate_span_processors(span_processors):
    """Validate that the span processor is a valid function."""
    span_processors = span_processors or []

    for span_processor in span_processors:
        if not callable(span_processor):
            raise MlflowException.invalid_parameter_value(
                "Span processor must be a callable function."
            )

        # Skip validation for builtin functions and partial functions that don't have __code__
        if not hasattr(span_processor, "__code__"):
            continue

        if span_processor.__code__.co_argcount != 1:
            raise MlflowException.invalid_parameter_value(
                "Span processor must take exactly one argument that accepts a LiveSpan object."
            )

    return span_processors



from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional

from mlflow.entities._mlflow_object import _MlflowObject
from mlflow.exceptions import MlflowException
from mlflow.protos import service_pb2 as pb


@dataclass
class MlflowExperimentLocation(_MlflowObject):
    """
    Represents the location of an MLflow experiment.

    Args:
        experiment_id: The ID of the MLflow experiment where the trace is stored.
    """

    experiment_id: str

    def to_proto(self):
        return pb.TraceLocation.MlflowExperimentLocation(experiment_id=self.experiment_id)

    @classmethod
    def from_proto(cls, proto) -> "MlflowExperimentLocation":
        return cls(experiment_id=proto.experiment_id)

    def to_dict(self) -> dict[str, Any]:
        return {"experiment_id": self.experiment_id}

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "MlflowExperimentLocation":
        return cls(experiment_id=d["experiment_id"])


@dataclass
class InferenceTableLocation(_MlflowObject):
    """
    Represents the location of a Databricks inference table.

    Args:
        full_table_name: The fully qualified name of the inference table where
            the trace is stored, in the format of `<catalog>.<schema>.<table>`.
    """

    full_table_name: str

    def to_proto(self):
        return pb.TraceLocation.InferenceTableLocation(full_table_name=self.full_table_name)

    @classmethod
    def from_proto(cls, proto) -> "InferenceTableLocation":
        return cls(full_table_name=proto.full_table_name)

    def to_dict(self) -> dict[str, Any]:
        return {"full_table_name": self.full_table_name}

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "InferenceTableLocation":
        return cls(full_table_name=d["full_table_name"])


class TraceLocationType(str, Enum):
    TRACE_LOCATION_TYPE_UNSPECIFIED = "TRACE_LOCATION_TYPE_UNSPECIFIED"
    MLFLOW_EXPERIMENT = "MLFLOW_EXPERIMENT"
    INFERENCE_TABLE = "INFERENCE_TABLE"

    def to_proto(self):
        return pb.TraceLocation.TraceLocationType.Value(self)

    @classmethod
    def from_proto(cls, proto: int) -> "TraceLocationType":
        return TraceLocationType(pb.TraceLocation.TraceLocationType.Name(proto))

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "TraceLocationType":
        return cls(d["type"])


@dataclass
class TraceLocation(_MlflowObject):
    """
    Represents the location where the trace is stored.

    Currently, MLflow supports two types of trace locations:

        - MLflow experiment: The trace is stored in an MLflow experiment.
        - Inference table: The trace is stored in a Databricks inference table.

    Args:
        type: The type of the trace location, should be one of the
            :py:class:`TraceLocationType` enum values.
        mlflow_experiment: The MLflow experiment location. Set this when the
            location type is MLflow experiment.
        inference_table: The inference table location. Set this when the
            location type is Databricks Inference table.
    """

    type: TraceLocationType
    mlflow_experiment: Optional[MlflowExperimentLocation] = None
    inference_table: Optional[InferenceTableLocation] = None

    def __post_init__(self) -> None:
        if self.mlflow_experiment is not None and self.inference_table is not None:
            raise MlflowException.invalid_parameter_value(
                "Only one of mlflow_experiment or inference_table can be provided."
            )

        if (self.mlflow_experiment and self.type != TraceLocationType.MLFLOW_EXPERIMENT) or (
            self.inference_table and self.type != TraceLocationType.INFERENCE_TABLE
        ):
            raise MlflowException.invalid_parameter_value(
                f"Trace location type {type} does not match the provided location "
                f"{self.mlflow_experiment or self.inference_table}."
            )

    def to_dict(self) -> dict[str, Any]:
        d = {"type": self.type.value}
        if self.mlflow_experiment:
            d["mlflow_experiment"] = self.mlflow_experiment.to_dict()
        elif self.inference_table:
            d["inference_table"] = self.inference_table.to_dict()
        return d

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "TraceLocation":
        return cls(
            type=TraceLocationType(d["type"]),
            mlflow_experiment=(
                MlflowExperimentLocation.from_dict(v) if (v := d.get("mlflow_experiment")) else None
            ),
            inference_table=(
                InferenceTableLocation.from_dict(v) if (v := d.get("inference_table")) else None
            ),
        )

    def to_proto(self):
        if self.mlflow_experiment:
            return pb.TraceLocation(
                type=self.type.to_proto(),
                mlflow_experiment=self.mlflow_experiment.to_proto(),
            )
        elif self.inference_table:
            return pb.TraceLocation(
                type=self.type.to_proto(),
                inference_table=self.inference_table.to_proto(),
            )

        else:
            return pb.TraceLocation(type=self.type.to_proto())

    @classmethod
    def from_proto(cls, proto) -> "TraceLocation":
        type_ = TraceLocationType.from_proto(proto.type)
        if proto.WhichOneof("identifier") == "mlflow_experiment":
            return cls(
                type=type_,
                mlflow_experiment=MlflowExperimentLocation.from_proto(proto.mlflow_experiment),
            )
        elif proto.WhichOneof("identifier") == "inference_table":
            return cls(
                type=type_,
                inference_table=InferenceTableLocation.from_proto(proto.inference_table),
            )
        else:
            return cls(type=type_)

    @classmethod
    def from_experiment_id(cls, experiment_id: str) -> "TraceLocation":
        return cls(
            type=TraceLocationType.MLFLOW_EXPERIMENT,
            mlflow_experiment=MlflowExperimentLocation(experiment_id=experiment_id),
        )



# Copyright 2018 Databricks, Inc.
import importlib.metadata
import re

VERSION = "3.1.2.dev0"


def is_release_version():
    return bool(re.match(r"^\d+\.\d+\.\d+$", VERSION))


def _is_package_installed(package_name: str) -> bool:
    try:
        importlib.metadata.version(package_name)
        return True
    except importlib.metadata.PackageNotFoundError:
        return False


# A flag to indicate whether the environment only has the tracing SDK
# installed, or includes the full MLflow or mlflow-skinny package.
# This is used to determine whether to import modules that require
# dependencies that are not included in the tracing SDK.
IS_TRACING_SDK_ONLY = not any(_is_package_installed(pkg) for pkg in ["mlflow", "mlflow-skinny"])



import base64
import json
import logging
import random
import time
import warnings
from functools import lru_cache
from typing import Any, Callable

import requests

from mlflow.environment_variables import (
    _MLFLOW_HTTP_REQUEST_MAX_BACKOFF_FACTOR_LIMIT,
    _MLFLOW_HTTP_REQUEST_MAX_RETRIES_LIMIT,
    MLFLOW_DATABRICKS_ENDPOINT_HTTP_RETRY_TIMEOUT,
    MLFLOW_ENABLE_DB_SDK,
    MLFLOW_HTTP_REQUEST_BACKOFF_FACTOR,
    MLFLOW_HTTP_REQUEST_BACKOFF_JITTER,
    MLFLOW_HTTP_REQUEST_MAX_RETRIES,
    MLFLOW_HTTP_REQUEST_TIMEOUT,
    MLFLOW_HTTP_RESPECT_RETRY_AFTER_HEADER,
)
from mlflow.exceptions import (
    CUSTOMER_UNAUTHORIZED,
    ERROR_CODE_TO_HTTP_STATUS,
    INVALID_PARAMETER_VALUE,
    InvalidUrlException,
    MlflowException,
    RestException,
    get_error_code,
)
from mlflow.protos import databricks_pb2
from mlflow.protos.databricks_pb2 import ENDPOINT_NOT_FOUND, ErrorCode
from mlflow.utils.proto_json_utils import parse_dict
from mlflow.utils.request_utils import (
    _TRANSIENT_FAILURE_RESPONSE_CODES,
    _get_http_response_with_retries,
    augmented_raise_for_status,  # noqa: F401
    cloud_storage_http_request,  # noqa: F401
)
from mlflow.utils.string_utils import strip_suffix

_logger = logging.getLogger(__name__)

RESOURCE_NON_EXISTENT = "RESOURCE_DOES_NOT_EXIST"
_REST_API_PATH_PREFIX = "/api/2.0"
_UC_OSS_REST_API_PATH_PREFIX = "/api/2.1"
_TRACE_REST_API_PATH_PREFIX = f"{_REST_API_PATH_PREFIX}/mlflow/traces"
_V3_REST_API_PATH_PREFIX = "/api/3.0"
_V3_TRACE_REST_API_PATH_PREFIX = f"{_V3_REST_API_PATH_PREFIX}/mlflow/traces"
_ARMERIA_OK = "200 OK"
_DATABRICKS_SDK_RETRY_AFTER_SECS_DEPRECATION_WARNING = (
    "The 'retry_after_secs' parameter of DatabricksError is deprecated"
)


def http_request(
    host_creds,
    endpoint,
    method,
    max_retries=None,
    backoff_factor=None,
    backoff_jitter=None,
    extra_headers=None,
    retry_codes=_TRANSIENT_FAILURE_RESPONSE_CODES,
    timeout=None,
    raise_on_status=True,
    respect_retry_after_header=None,
    retry_timeout_seconds=None,
    **kwargs,
):
    """Makes an HTTP request with the specified method to the specified hostname/endpoint. Transient
    errors such as Rate-limited (429), service unavailable (503) and internal error (500) are
    retried with an exponential back off with backoff_factor * (1, 2, 4, ... seconds).
    The function parses the API response (assumed to be JSON) into a Python object and returns it.

    Args:
        host_creds: A :py:class:`mlflow.rest_utils.MlflowHostCreds` object containing
            hostname and optional authentication.
        endpoint: A string for service endpoint, e.g. "/path/to/object".
        method: A string indicating the method to use, e.g. "GET", "POST", "PUT".
        max_retries: Maximum number of retries before throwing an exception.
        backoff_factor: A time factor for exponential backoff. e.g. value 5 means the HTTP
            request will be retried with interval 5, 10, 20... seconds. A value of 0 turns off the
            exponential backoff.
        backoff_jitter: A random jitter to add to the backoff interval.
        extra_headers: A dict of HTTP header name-value pairs to be included in the request.
        retry_codes: A list of HTTP response error codes that qualifies for retry.
        timeout: Wait for timeout seconds for response from remote server for connect and
            read request.
        raise_on_status: Whether to raise an exception, or return a response, if status falls
            in retry_codes range and retries have been exhausted.
        respect_retry_after_header: Whether to respect Retry-After header on status codes defined
            as Retry.RETRY_AFTER_STATUS_CODES or not.
        retry_timeout_seconds: Timeout for retires. Only effective when using Databricks SDK.
        kwargs: Additional keyword arguments to pass to `requests.Session.request()`

    Returns:
        requests.Response object.
    """
    cleaned_hostname = strip_suffix(host_creds.host, "/")
    url = f"{cleaned_hostname}{endpoint}"

    # Set defaults for retry parameters from environment variables if not specified
    max_retries = MLFLOW_HTTP_REQUEST_MAX_RETRIES.get() if max_retries is None else max_retries
    backoff_factor = (
        MLFLOW_HTTP_REQUEST_BACKOFF_FACTOR.get() if backoff_factor is None else backoff_factor
    )
    backoff_jitter = (
        MLFLOW_HTTP_REQUEST_BACKOFF_JITTER.get() if backoff_jitter is None else backoff_jitter
    )

    if host_creds.use_databricks_sdk:
        from databricks.sdk.errors import DatabricksError

        ws_client = get_workspace_client(
            host_creds.use_secret_scope_token,
            host_creds.host,
            host_creds.token,
            host_creds.databricks_auth_profile,
            retry_timeout_seconds=retry_timeout_seconds,
        )

        def make_sdk_call():
            # Databricks SDK `APIClient.do` API is for making request using
            # HTTP
            # https://github.com/databricks/databricks-sdk-py/blob/a714146d9c155dd1e3567475be78623f72028ee0/databricks/sdk/core.py#L134
            # suppress the warning due to https://github.com/databricks/databricks-sdk-py/issues/963
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", message=f".*{_DATABRICKS_SDK_RETRY_AFTER_SECS_DEPRECATION_WARNING}.*"
                )
                raw_response = ws_client.api_client.do(
                    method=method,
                    path=endpoint,
                    headers=extra_headers,
                    raw=True,
                    query=kwargs.get("params"),
                    body=kwargs.get("json"),
                    files=kwargs.get("files"),
                    data=kwargs.get("data"),
                )
                return raw_response["contents"]._response

        try:
            # We retry the SDK call with exponential backoff because the Databricks SDK default
            # retry behavior does not handle all transient errors that we want to retry, and it
            # does not support a customizable retry policy based on HTTP response status codes.
            # Note that, in uncommon cases (due to the limited set if HTTP status codes and
            # response strings that Databricks SDK retries on), the SDK may retry internally,
            # and MLflow may retry on top of that, leading to more retries than specified by
            # `max_retries`. This is acceptable, given the enforcement of an overall request
            # timeout via `retry_timeout_seconds`.
            #
            # TODO: Update transient error handling defaults in Databricks SDK to match standard
            # practices (e.g. retrying on 429, 500, 503, etc.), support custom retries in Databricks
            # SDK, and remove this custom retry wrapper from MLflow
            return _retry_databricks_sdk_call_with_exponential_backoff(
                call_func=make_sdk_call,
                retry_codes=retry_codes,
                retry_timeout_seconds=(
                    retry_timeout_seconds
                    if retry_timeout_seconds is not None
                    else MLFLOW_DATABRICKS_ENDPOINT_HTTP_RETRY_TIMEOUT.get()
                ),
                backoff_factor=backoff_factor,
                backoff_jitter=backoff_jitter,
                max_retries=max_retries,
            )
        except DatabricksError as e:
            response = requests.Response()
            response.url = url
            response.status_code = ERROR_CODE_TO_HTTP_STATUS.get(e.error_code, 500)
            response.reason = str(e)
            response.encoding = "UTF-8"
            response._content = json.dumps(
                {
                    "error_code": e.error_code,
                    "message": str(e),
                }
            ).encode("UTF-8")
            return response

    _validate_max_retries(max_retries)
    _validate_backoff_factor(backoff_factor)
    respect_retry_after_header = (
        MLFLOW_HTTP_RESPECT_RETRY_AFTER_HEADER.get()
        if respect_retry_after_header is None
        else respect_retry_after_header
    )

    timeout = MLFLOW_HTTP_REQUEST_TIMEOUT.get() if timeout is None else timeout
    auth_str = None
    if host_creds.username and host_creds.password:
        basic_auth_str = f"{host_creds.username}:{host_creds.password}".encode()
        auth_str = "Basic " + base64.standard_b64encode(basic_auth_str).decode("utf-8")
    elif host_creds.token:
        auth_str = f"Bearer {host_creds.token}"
    elif host_creds.client_secret:
        raise MlflowException(
            "To use OAuth authentication, set environmental variable "
            f"'{MLFLOW_ENABLE_DB_SDK.name}' to true",
            error_code=CUSTOMER_UNAUTHORIZED,
        )

    from mlflow.tracking.request_header.registry import resolve_request_headers

    headers = dict(**resolve_request_headers())
    if extra_headers:
        headers = dict(**headers, **extra_headers)

    if auth_str:
        headers["Authorization"] = auth_str

    if host_creds.client_cert_path is not None:
        kwargs["cert"] = host_creds.client_cert_path

    if host_creds.aws_sigv4:
        # will overwrite the Authorization header
        from requests_auth_aws_sigv4 import AWSSigV4

        kwargs["auth"] = AWSSigV4("execute-api")
    elif host_creds.auth:
        from mlflow.tracking.request_auth.registry import fetch_auth

        kwargs["auth"] = fetch_auth(host_creds.auth)

    try:
        return _get_http_response_with_retries(
            method,
            url,
            max_retries,
            backoff_factor,
            backoff_jitter,
            retry_codes,
            raise_on_status,
            headers=headers,
            verify=host_creds.verify,
            timeout=timeout,
            respect_retry_after_header=respect_retry_after_header,
            **kwargs,
        )
    except requests.exceptions.Timeout as to:
        raise MlflowException(
            f"API request to {url} failed with timeout exception {to}."
            " To increase the timeout, set the environment variable "
            f"{MLFLOW_HTTP_REQUEST_TIMEOUT!s} to a larger value."
        ) from to
    except requests.exceptions.InvalidURL as iu:
        raise InvalidUrlException(f"Invalid url: {url}") from iu
    except Exception as e:
        raise MlflowException(f"API request to {url} failed with exception {e}")


@lru_cache(maxsize=5)
def get_workspace_client(
    use_secret_scope_token,
    host,
    token,
    databricks_auth_profile,
    retry_timeout_seconds=None,
):
    from databricks.sdk import WorkspaceClient
    from databricks.sdk.config import Config

    if use_secret_scope_token:
        kwargs = {"host": host, "token": token}
    else:
        kwargs = {"profile": databricks_auth_profile}
    config = Config(
        **kwargs,
        retry_timeout_seconds=retry_timeout_seconds
        or MLFLOW_DATABRICKS_ENDPOINT_HTTP_RETRY_TIMEOUT.get(),
    )
    # Note: If we use `config` param, all SDK configurations must be set in `config` object.
    return WorkspaceClient(config=config)


def _can_parse_as_json_object(string):
    try:
        return isinstance(json.loads(string), dict)
    except Exception:
        return False


def http_request_safe(host_creds, endpoint, method, **kwargs):
    """
    Wrapper around ``http_request`` that also verifies that the request succeeds with code 200.
    """
    response = http_request(host_creds=host_creds, endpoint=endpoint, method=method, **kwargs)
    return verify_rest_response(response, endpoint)


def verify_rest_response(response, endpoint):
    """Verify the return code and format, raise exception if the request was not successful."""
    # Handle Armeria-specific response case where response text is "200 OK"
    if response.status_code == 200 and response.text.strip() == _ARMERIA_OK:
        response._content = b"{}"  # Update response content to be an empty JSON dictionary
        return response

    # Handle non-200 status codes
    if response.status_code != 200:
        if _can_parse_as_json_object(response.text):
            raise RestException(json.loads(response.text))
        else:
            base_msg = (
                f"API request to endpoint {endpoint} "
                f"failed with error code {response.status_code} != 200"
            )
            raise MlflowException(
                f"{base_msg}. Response body: '{response.text}'",
                error_code=get_error_code(response.status_code),
            )

    # Skip validation for endpoints (e.g. DBFS file-download API) which may return a non-JSON
    # response
    if endpoint.startswith(_REST_API_PATH_PREFIX) and not _can_parse_as_json_object(response.text):
        base_msg = (
            "API request to endpoint was successful but the response body was not "
            "in a valid JSON format"
        )
        raise MlflowException(f"{base_msg}. Response body: '{response.text}'")

    return response


def _validate_max_retries(max_retries):
    max_retry_limit = _MLFLOW_HTTP_REQUEST_MAX_RETRIES_LIMIT.get()

    if max_retry_limit < 0:
        raise MlflowException(
            message=f"The current maximum retry limit is invalid ({max_retry_limit}). "
            "Cannot be negative.",
            error_code=INVALID_PARAMETER_VALUE,
        )
    if max_retries >= max_retry_limit:
        raise MlflowException(
            message=f"The configured max_retries value ({max_retries}) is "
            f"in excess of the maximum allowable retries ({max_retry_limit})",
            error_code=INVALID_PARAMETER_VALUE,
        )

    if max_retries < 0:
        raise MlflowException(
            message=f"The max_retries value must be either 0 a positive integer. Got {max_retries}",
            error_code=INVALID_PARAMETER_VALUE,
        )


def _validate_backoff_factor(backoff_factor):
    max_backoff_factor_limit = _MLFLOW_HTTP_REQUEST_MAX_BACKOFF_FACTOR_LIMIT.get()

    if max_backoff_factor_limit < 0:
        raise MlflowException(
            message="The current maximum backoff factor limit is invalid "
            f"({max_backoff_factor_limit}). Cannot be negative.",
            error_code=INVALID_PARAMETER_VALUE,
        )

    if backoff_factor >= max_backoff_factor_limit:
        raise MlflowException(
            message=f"The configured backoff_factor value ({backoff_factor}) is in excess "
            "of the maximum allowable backoff_factor limit "
            f"({max_backoff_factor_limit})",
            error_code=INVALID_PARAMETER_VALUE,
        )

    if backoff_factor < 0:
        raise MlflowException(
            message="The backoff_factor value must be either 0 a positive integer. "
            f"Got {backoff_factor}",
            error_code=INVALID_PARAMETER_VALUE,
        )


def _time_sleep(seconds: float) -> None:
    """
    This function is specifically mocked in `test_rest_utils.py` to test the backoff logic in
    isolation. We avoid wrapping `time.sleep` globally to prevent interfering with unrelated sleep
    calls elsewhere in the codebase or in third-party libraries.
    """
    time.sleep(seconds)


def _retry_databricks_sdk_call_with_exponential_backoff(
    *,
    call_func: Callable[..., Any],
    retry_codes: list[int],
    retry_timeout_seconds: int,
    backoff_factor: int,
    backoff_jitter: float,
    max_retries: int,
):
    """
    Retry a Databricks SDK call with exponential backoff until timeout or max retries reached.

    Args:
        call_func: Function to call that may raise DatabricksError
        retry_codes: Set of HTTP status codes that should trigger retries
        retry_timeout_seconds: Maximum time to spend retrying in seconds
        backoff_factor: Factor for exponential backoff
        backoff_jitter: Random jitter to add to backoff
        max_retries: Maximum number of retry attempts

    Returns:
        The result of call_func() on success

    Raises:
        DatabricksError: If all retries are exhausted or non-retryable error occurs
    """
    from databricks.sdk.errors import DatabricksError

    start_time = time.time()
    attempt = 0

    while attempt <= max_retries:
        try:
            return call_func()
        except DatabricksError as e:
            # Get HTTP status code from the error
            status_code = ERROR_CODE_TO_HTTP_STATUS.get(e.error_code, 500)

            # Check if this is a retryable error
            if status_code not in retry_codes:
                raise

            # Check if we've exceeded max retries
            if attempt >= max_retries:
                _logger.warning(f"Max retries ({max_retries}) exceeded: {e}")
                raise

            # Calculate backoff time with exponential backoff and jitter
            # NB: Ideally, we'd use urllib3.Retry to compute the jitter, check whether we've
            # exceed max retries, etc. However, urllib3.Retry in urllib3 version 1.x, which MLflow
            # maintains compatibility with, doesn't support retries with jitter
            if attempt <= 0:
                backoff_time = 0  # No backoff on first retry attempt
            else:
                backoff_time = backoff_factor * (2**attempt)
                if backoff_jitter > 0:
                    backoff_time += random.random() * backoff_jitter

            # Check if we've exceeded or would exceed timeout
            elapsed_time = time.time() - start_time
            if elapsed_time + backoff_time >= retry_timeout_seconds:
                _logger.warning(f"Retry timeout ({retry_timeout_seconds}s) exceeded: {e}")
                raise

            _logger.debug(
                f"Databricks SDK call failed with retryable error (status {status_code}): {e}. "
                f"Retrying in {backoff_time:.2f} seconds (attempt {attempt + 1})"
            )

            _time_sleep(backoff_time)
            attempt += 1


def _get_path(path_prefix, endpoint_path):
    return f"{path_prefix}{endpoint_path}"


def extract_api_info_for_service(service, path_prefix):
    """Return a dictionary mapping each API method to a tuple (path, HTTP method)"""
    service_methods = service.DESCRIPTOR.methods
    res = {}
    for service_method in service_methods:
        endpoints = service_method.GetOptions().Extensions[databricks_pb2.rpc].endpoints
        endpoint = endpoints[0]
        endpoint_path = _get_path(path_prefix, endpoint.path)
        res[service().GetRequestClass(service_method)] = (endpoint_path, endpoint.method)
    return res


def extract_all_api_info_for_service(service, path_prefix):
    """Return a dictionary mapping each API method to a list of tuples [(path, HTTP method)]"""
    service_methods = service.DESCRIPTOR.methods
    res = {}
    for service_method in service_methods:
        endpoints = service_method.GetOptions().Extensions[databricks_pb2.rpc].endpoints
        res[service().GetRequestClass(service_method)] = [
            (_get_path(path_prefix, endpoint.path), endpoint.method) for endpoint in endpoints
        ]
    return res


def get_single_trace_endpoint(request_id, use_v3=True):
    """
    Get the endpoint for a single trace.
    For Databricks tracking URIs, use the V3 API.
    For all other tracking URIs, use the V2 API.

    Args:
        request_id: The trace ID.
        use_v3: Whether to use the V3 API. If True, use the V3 API. If False, use the V2 API.
    """
    if use_v3:
        return f"{_V3_TRACE_REST_API_PATH_PREFIX}/{request_id}"
    return f"{_TRACE_REST_API_PATH_PREFIX}/{request_id}"


def get_logged_model_endpoint(model_id: str) -> str:
    return f"{_REST_API_PATH_PREFIX}/mlflow/logged-models/{model_id}"


def get_single_assessment_endpoint(trace_id: str, assessment_id: str) -> str:
    """
    Get the endpoint for a single assessment.

    Args:
        trace_id: The trace ID.
        assessment_id: The assessment ID.
    """
    return f"{_V3_TRACE_REST_API_PATH_PREFIX}/{trace_id}/assessments/{assessment_id}"


def get_trace_tag_endpoint(trace_id):
    """Get the endpoint for trace tags. Always use v2 endpoint."""
    return f"{_REST_API_PATH_PREFIX}/mlflow/traces/{trace_id}/tags"


def call_endpoint(
    host_creds,
    endpoint,
    method,
    json_body,
    response_proto,
    extra_headers=None,
    retry_timeout_seconds=None,
):
    # Convert json string to json dictionary, to pass to requests
    if json_body is not None:
        json_body = json.loads(json_body)
    call_kwargs = {
        "host_creds": host_creds,
        "endpoint": endpoint,
        "method": method,
    }
    if extra_headers is not None:
        call_kwargs["extra_headers"] = extra_headers
    if retry_timeout_seconds is not None:
        call_kwargs["retry_timeout_seconds"] = retry_timeout_seconds
    if method == "GET":
        call_kwargs["params"] = json_body
        response = http_request(**call_kwargs)
    else:
        call_kwargs["json"] = json_body
        response = http_request(**call_kwargs)

    response = verify_rest_response(response, endpoint)
    response_to_parse = response.text
    try:
        js_dict = json.loads(response_to_parse)
    except json.JSONDecodeError:
        if len(response_to_parse) > 50:
            response_to_parse = response_to_parse[:50] + "..."
        _logger.warning(f"Response is not a valid JSON object: {response_to_parse}")
        raise

    parse_dict(js_dict=js_dict, message=response_proto)
    return response_proto


def call_endpoints(host_creds, endpoints, json_body, response_proto, extra_headers=None):
    # The order that the endpoints are called in is defined by the order
    # specified in ModelRegistryService in model_registry.proto
    for i, (endpoint, method) in enumerate(endpoints):
        try:
            return call_endpoint(
                host_creds, endpoint, method, json_body, response_proto, extra_headers
            )
        except RestException as e:
            if e.error_code != ErrorCode.Name(ENDPOINT_NOT_FOUND) or i == len(endpoints) - 1:
                raise e


class MlflowHostCreds:
    """
    Provides a hostname and optional authentication for talking to an MLflow tracking server.

    Args:
        host: Hostname (e.g., http://localhost:5000) to MLflow server. Required.
        username: Username to use with Basic authentication when talking to server.
            If this is specified, password must also be specified.
        password: Password to use with Basic authentication when talking to server.
            If this is specified, username must also be specified.
        token: Token to use with Bearer authentication when talking to server.
            If provided, user/password authentication will be ignored.
        aws_sigv4: If true, we will create a signature V4 to be added for any outgoing request.
            Keys for signing the request can be passed via ENV variables,
            or will be fetched via boto3 session.
        auth: If set, the auth will be added for any outgoing request.
            Keys for signing the request can be passed via ENV variables,
        ignore_tls_verification: If true, we will not verify the server's hostname or TLS
            certificate. This is useful for certain testing situations, but should never be
            true in production.
            If this is set to true ``server_cert_path`` must not be set.
        client_cert_path: Path to ssl client cert file (.pem).
            Sets the cert param of the ``requests.request``
            function (see https://requests.readthedocs.io/en/master/api/).
        server_cert_path: Path to a CA bundle to use.
            Sets the verify param of the ``requests.request``
            function (see https://requests.readthedocs.io/en/master/api/).
            If this is set ``ignore_tls_verification`` must be false.
        use_databricks_sdk: A boolean value represent whether using Databricks SDK for
            authentication.
        databricks_auth_profile: The name of the profile used by Databricks SDK for
            authentication.
        client_id: The client ID used by Databricks OAuth
        client_secret: The client secret used by Databricks OAuth
    """

    def __init__(
        self,
        host,
        username=None,
        password=None,
        token=None,
        aws_sigv4=False,
        auth=None,
        ignore_tls_verification=False,
        client_cert_path=None,
        server_cert_path=None,
        use_databricks_sdk=False,
        databricks_auth_profile=None,
        client_id=None,
        client_secret=None,
        use_secret_scope_token=False,
    ):
        if not host:
            raise MlflowException(
                message="host is a required parameter for MlflowHostCreds",
                error_code=INVALID_PARAMETER_VALUE,
            )
        if ignore_tls_verification and (server_cert_path is not None):
            raise MlflowException(
                message=(
                    "When 'ignore_tls_verification' is true then 'server_cert_path' "
                    "must not be set! This error may have occurred because the "
                    "'MLFLOW_TRACKING_INSECURE_TLS' and 'MLFLOW_TRACKING_SERVER_CERT_PATH' "
                    "environment variables are both set - only one of these environment "
                    "variables may be set."
                ),
                error_code=INVALID_PARAMETER_VALUE,
            )
        self.host = host
        self.username = username
        self.password = password
        self.token = token
        self.aws_sigv4 = aws_sigv4
        self.auth = auth
        self.ignore_tls_verification = ignore_tls_verification
        self.client_cert_path = client_cert_path
        self.server_cert_path = server_cert_path
        self.use_databricks_sdk = use_databricks_sdk
        self.databricks_auth_profile = databricks_auth_profile
        self.client_id = client_id
        self.client_secret = client_secret
        self.use_secret_scope_token = use_secret_scope_token

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    @property
    def verify(self):
        if self.use_databricks_sdk:
            # Let databricks-sdk set HTTP request `verify` param.
            return None
        if self.server_cert_path is None:
            return not self.ignore_tls_verification
        else:
            return self.server_cert_path



import inspect
import logging
from typing import Any, Generator, Optional

from mlflow.exceptions import MlflowException
from mlflow.models.utils import _convert_llm_ndarray_to_list
from mlflow.protos.databricks_pb2 import INTERNAL_ERROR
from mlflow.pyfunc.model import (
    _load_context_model_and_signature,
)
from mlflow.types.llm import ChatCompletionChunk, ChatCompletionResponse, ChatMessage, ChatParams

_logger = logging.getLogger(__name__)


def _load_pyfunc(model_path: str, model_config: Optional[dict[str, Any]] = None):
    context, chat_model, signature = _load_context_model_and_signature(model_path, model_config)
    return _ChatModelPyfuncWrapper(chat_model=chat_model, context=context, signature=signature)


class _ChatModelPyfuncWrapper:
    """
    Wrapper class that converts dict inputs to pydantic objects accepted by :class:`~ChatModel`.
    """

    def __init__(self, chat_model, context, signature):
        """
        Args:
            chat_model: An instance of a subclass of :class:`~ChatModel`.
            context: A :class:`~PythonModelContext` instance containing artifacts that
                        ``chat_model`` may use when performing inference.
            signature: :class:`~ModelSignature` instance describing model input and output.
        """
        self.chat_model = chat_model
        self.context = context
        self.signature = signature

    def get_raw_model(self):
        """
        Returns the underlying model.
        """
        return self.chat_model

    def _convert_input(self, model_input):
        import pandas

        if isinstance(model_input, dict):
            dict_input = model_input
        elif isinstance(model_input, pandas.DataFrame):
            dict_input = {
                k: _convert_llm_ndarray_to_list(v[0])
                for k, v in model_input.to_dict(orient="list").items()
            }
        else:
            raise MlflowException(
                "Unsupported model input type. Expected a dict or pandas.DataFrame, "
                f"but got {type(model_input)} instead.",
                error_code=INTERNAL_ERROR,
            )

        messages = [ChatMessage.from_dict(message) for message in dict_input.pop("messages", [])]
        params = ChatParams.from_dict(dict_input)
        return messages, params

    def predict(
        self, model_input: dict[str, Any], params: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        """
        Args:
            model_input: Model input data in the form of a chat request.
            params: Additional parameters to pass to the model for inference.
                       Unused in this implementation, as the params are handled
                       via ``self._convert_input()``.

        Returns:
            Model predictions in :py:class:`~ChatCompletionResponse` format.
        """
        messages, params = self._convert_input(model_input)
        parameters = inspect.signature(self.chat_model.predict).parameters
        if "context" in parameters or len(parameters) == 3:
            response = self.chat_model.predict(self.context, messages, params)
        else:
            response = self.chat_model.predict(messages, params)
        return self._response_to_dict(response)

    def _response_to_dict(self, response: ChatCompletionResponse) -> dict[str, Any]:
        if not isinstance(response, ChatCompletionResponse):
            raise MlflowException(
                "Model returned an invalid response. Expected a ChatCompletionResponse, but "
                f"got {type(response)} instead.",
                error_code=INTERNAL_ERROR,
            )
        return response.to_dict()

    def _streaming_response_to_dict(self, response: ChatCompletionChunk) -> dict[str, Any]:
        if not isinstance(response, ChatCompletionChunk):
            raise MlflowException(
                "Model returned an invalid response. Expected a ChatCompletionChunk, but "
                f"got {type(response)} instead.",
                error_code=INTERNAL_ERROR,
            )
        return response.to_dict()

    def predict_stream(
        self, model_input: dict[str, Any], params: Optional[dict[str, Any]] = None
    ) -> Generator[dict[str, Any], None, None]:
        """
        Args:
            model_input: Model input data in the form of a chat request.
            params: Additional parameters to pass to the model for inference.
                       Unused in this implementation, as the params are handled
                       via ``self._convert_input()``.

        Returns:
            Generator over model predictions in :py:class:`~ChatCompletionChunk` format.
        """
        messages, params = self._convert_input(model_input)
        parameters = inspect.signature(self.chat_model.predict_stream).parameters
        if "context" in parameters or len(parameters) == 3:
            stream = self.chat_model.predict_stream(self.context, messages, params)
        else:
            stream = self.chat_model.predict_stream(messages, params)

        for response in stream:
            yield self._streaming_response_to_dict(response)



import functools
import logging
import os
import time
from typing import Optional

from mlflow.environment_variables import _MLFLOW_TESTING
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import RESOURCE_DOES_NOT_EXIST

_logger = logging.getLogger(__name__)


# NB: The maxsize=1 is added for encouraging the cache refresh so the user doesn't get stale
#    commit hash from the cache. This doesn't work perfectly because it only updates cache
#    when the user calls it with a different repo name, but it's better than nothing.
@functools.lru_cache(maxsize=1)
def get_latest_commit_for_repo(repo: str) -> str:
    """
    Fetches the latest commit hash for a repository from the HuggingFace model hub.
    """
    try:
        import huggingface_hub as hub
    except ImportError:
        raise MlflowException(
            "Unable to fetch model commit hash from the HuggingFace model hub. "
            "This is required for saving Transformer model without base model "
            "weights, while ensuring the version consistency of the model. "
            "Please install the `huggingface-hub` package and retry.",
            error_code=RESOURCE_DOES_NOT_EXIST,
        )

    from huggingface_hub.errors import HfHubHTTPError

    api = hub.HfApi()
    for i in range(7):
        try:
            return api.model_info(repo).sha
        except HfHubHTTPError as e:
            if not _MLFLOW_TESTING.get():
                raise

            # Retry on rate limit error
            if e.response.status_code == 429:
                _logger.warning(
                    f"Rate limit exceeded while fetching commit hash for repo {repo}. "
                    f"Retrying in {2**i} seconds. Error: {e}",
                )
                time.sleep(2**i)
                continue
            raise

    raise MlflowException(
        "Unable to fetch model commit hash from the HuggingFace model hub. "
        "This is required for saving Transformer model without base model "
        "weights, while ensuring the version consistency of the model. ",
        error_code=RESOURCE_DOES_NOT_EXIST,
    )


def is_valid_hf_repo_id(maybe_repo_id: Optional[str]) -> bool:
    """
    Check if the given string is a valid HuggingFace repo identifier e.g. "username/repo_id".
    """

    if not maybe_repo_id or os.path.isdir(maybe_repo_id):
        return False

    try:
        from huggingface_hub.utils import HFValidationError, validate_repo_id
    except ImportError:
        raise MlflowException(
            "Unable to validate the repository identifier for the HuggingFace model hub "
            "because the `huggingface-hub` package is not installed. Please install the "
            "package with `pip install huggingface-hub` command and retry."
        )

    try:
        validate_repo_id(maybe_repo_id)
        return True
    except HFValidationError as e:
        _logger.warning(f"The repository identified {maybe_repo_id} is invalid: {e}")
        return False



import logging
import os
from functools import lru_cache

from mlflow.tracking.context.git_context import GitRunContext
from mlflow.tracking.context.registry import resolve_tags
from mlflow.utils.databricks_utils import is_in_databricks_notebook
from mlflow.utils.git_utils import get_git_branch, get_git_commit, get_git_repo_url
from mlflow.utils.mlflow_tags import (
    MLFLOW_GIT_BRANCH,
    MLFLOW_GIT_COMMIT,
    MLFLOW_GIT_REPO_URL,
    TRACE_RESOLVE_TAGS_ALLOWLIST,
)

_logger = logging.getLogger(__name__)


@lru_cache(maxsize=1)
def resolve_env_metadata():
    """
    Resolve common environment metadata to be saved in the trace info. These should not
    # change over time, so we resolve them only once. These will be stored in trace
    # metadata rather than tags, because they are immutable.
    """
    # GitRunContext does not property work in notebook because _get_main_file()
    # points to the kernel launcher file, not the actual notebook file.
    metadata = resolve_tags(ignore=[GitRunContext])
    if not is_in_databricks_notebook():
        # Get Git metadata for the script or notebook. If the notebook is in a
        # Databricks managed Git repo, DatabricksRepoRunContext the metadata
        # so we don't need to run this logic.
        metadata.update(_resolve_git_metadata())

    return {key: value for key, value in metadata.items() if key in TRACE_RESOLVE_TAGS_ALLOWLIST}


def _resolve_git_metadata():
    try:
        import git  # noqa: F401
    except ImportError:
        _logger.debug("Git python package is not installed. Skipping git metadata resolution.")
        return {}

    try:
        repo = os.getcwd()
        return {
            MLFLOW_GIT_COMMIT: get_git_commit(repo) or "",
            MLFLOW_GIT_REPO_URL: get_git_repo_url(repo) or "",
            MLFLOW_GIT_BRANCH: get_git_branch(repo) or "",
        }
    except Exception:
        _logger.debug("Failed to resolve git metadata", exc_info=True)

    return {}



import logging
import operator
import os
from decimal import Decimal
from typing import Optional

from mlflow.exceptions import MlflowException
from mlflow.models.evaluation import EvaluationResult
from mlflow.protos.databricks_pb2 import BAD_REQUEST, INVALID_PARAMETER_VALUE

_logger = logging.getLogger(__name__)


class MetricThreshold:
    """
    This class allows you to define metric thresholds for model validation.
    Allowed thresholds are: threshold, min_absolute_change, min_relative_change.

    Args:
        threshold: (Optional) A number representing the value threshold for the metric.

            - If greater is better for the metric, the metric value has to be
              >= threshold to pass validation.
            - Otherwise, the metric value has to be <= threshold to pass the validation.

        min_absolute_change: (Optional) A positive number representing the minimum absolute
            change required for candidate model to pass validation with
            the baseline model.

            - If greater is better for the metric, metric value has to be
              >= baseline model metric value + min_absolute_change to pass the validation.
            - Otherwise, metric value has to be <= baseline model metric value - min_absolute_change
              to pass the validation.

        min_relative_change: (Optional) A floating point number between 0 and 1 representing
            the minimum relative change (in percentage of
            baseline model metric value) for candidate model
            to pass the comparison with the baseline model.

            - If greater is better for the metric, metric value has to be
              >= baseline model metric value * (1 + min_relative_change)
            - Otherwise, metric value has to be
              <= baseline model metric value * (1 - min_relative_change)
            - Note that if the baseline model metric value is equal to 0, the
              threshold falls back performing a simple verification that the
              candidate metric value is better than the baseline metric value,
              i.e. metric value >= baseline model metric value + 1e-10 if greater
              is better; metric value <= baseline model metric value - 1e-10 if
              lower is better.

        greater_is_better: A required boolean representing whether greater value is
            better for the metric.
    """

    def __init__(
        self,
        threshold=None,
        min_absolute_change=None,
        min_relative_change=None,
        greater_is_better=None,
    ):
        if threshold is not None and type(threshold) not in {int, float}:
            raise MetricThresholdClassException("`threshold` parameter must be a number.")
        if min_absolute_change is not None and (
            type(min_absolute_change) not in {int, float} or min_absolute_change <= 0
        ):
            raise MetricThresholdClassException(
                "`min_absolute_change` parameter must be a positive number."
            )
        if min_relative_change is not None:
            if not isinstance(min_relative_change, float):
                raise MetricThresholdClassException(
                    "`min_relative_change` parameter must be a floating point number."
                )
            if min_relative_change < 0 or min_relative_change > 1:
                raise MetricThresholdClassException(
                    "`min_relative_change` parameter must be between 0 and 1."
                )
        if greater_is_better is None:
            raise MetricThresholdClassException("`greater_is_better` parameter must be defined.")
        if not isinstance(greater_is_better, bool):
            raise MetricThresholdClassException("`greater_is_better` parameter must be a boolean.")
        if threshold is None and min_absolute_change is None and min_relative_change is None:
            raise MetricThresholdClassException("no threshold was specified.")
        self._threshold = threshold
        self._min_absolute_change = min_absolute_change
        self._min_relative_change = min_relative_change
        self._greater_is_better = greater_is_better

    @property
    def threshold(self):
        """
        Value of the threshold.
        """
        return self._threshold

    @property
    def min_absolute_change(self):
        """
        Value of the minimum absolute change required to pass model comparison with baseline model.
        """
        return self._min_absolute_change

    @property
    def min_relative_change(self):
        """
        Float value of the minimum relative change required to pass model comparison with
        baseline model.
        """
        return self._min_relative_change

    @property
    def greater_is_better(self):
        """
        Boolean value representing whether greater value is better for the metric.
        """
        return self._greater_is_better

    def __str__(self):
        """
        Returns a human-readable string consisting of all specified thresholds.
        """
        threshold_strs = []
        if self._threshold is not None:
            threshold_strs.append(f"Threshold: {self._threshold}.")
        if self._min_absolute_change is not None:
            threshold_strs.append(f"Minimum Absolute Change: {self._min_absolute_change}.")
        if self._min_relative_change is not None:
            threshold_strs.append(f"Minimum Relative Change: {self._min_relative_change}.")
        if self._greater_is_better is not None:
            if self._greater_is_better:
                threshold_strs.append("Greater value is better.")
            else:
                threshold_strs.append("Lower value is better.")
        return " ".join(threshold_strs)


class MetricThresholdClassException(MlflowException):
    def __init__(self, _message, **kwargs):
        message = "Could not instantiate MetricThreshold class: " + _message
        super().__init__(message, error_code=INVALID_PARAMETER_VALUE, **kwargs)


class _MetricValidationResult:
    """
    Internal class for representing validation result per metric.
    Not user facing, used for organizing metric failures and generating failure message
    more conveniently.

    Args:
        metric_name: String representing the metric name
        candidate_metric_value: value of metric for candidate model
        metric_threshold: :py:class: `MetricThreshold<mlflow.models.validation.MetricThreshold>`
            The MetricThreshold for the metric.
        baseline_metric_value: value of metric for baseline model
    """

    missing_candidate = False
    missing_baseline = False
    threshold_failed = False
    min_absolute_change_failed = False
    min_relative_change_failed = False

    def __init__(
        self,
        metric_name,
        candidate_metric_value,
        metric_threshold,
        baseline_metric_value=None,
    ):
        self.metric_name = metric_name
        self.candidate_metric_value = candidate_metric_value
        self.baseline_metric_value = baseline_metric_value
        self.metric_threshold = metric_threshold

    def __str__(self):
        """
        Returns a human-readable string representing the validation result for the metric.
        """
        if self.is_success():
            return f"Metric {self.metric_name} passed the validation."

        if self.missing_candidate:
            return (
                f"Metric validation failed: metric {self.metric_name} was missing from the "
                f"evaluation result of the candidate model."
            )

        result_strs = []
        if self.threshold_failed:
            result_strs.append(
                f"Metric {self.metric_name} value threshold check failed: "
                f"candidate model {self.metric_name} = {self.candidate_metric_value}, "
                f"{self.metric_name} threshold = {self.metric_threshold.threshold}."
            )
        if self.missing_baseline:
            result_strs.append(
                f"Model comparison failed: metric {self.metric_name} was missing from "
                f"the evaluation result of the baseline model."
            )
        else:
            if self.min_absolute_change_failed:
                result_strs.append(
                    f"Metric {self.metric_name} minimum absolute change check failed: "
                    f"candidate model {self.metric_name} = {self.candidate_metric_value}, "
                    f"baseline model {self.metric_name} = {self.baseline_metric_value}, "
                    f"{self.metric_name} minimum absolute change threshold = "
                    f"{self.metric_threshold.min_absolute_change}."
                )
            if self.min_relative_change_failed:
                result_strs.append(
                    f"Metric {self.metric_name} minimum relative change check failed: "
                    f"candidate model {self.metric_name} = {self.candidate_metric_value}, "
                    f"baseline model {self.metric_name} = {self.baseline_metric_value}, "
                    f"{self.metric_name} minimum relative change threshold = "
                    f"{self.metric_threshold.min_relative_change}."
                )
        return " ".join(result_strs)

    def is_success(self):
        return (
            not self.missing_candidate
            and not self.missing_baseline
            and not self.threshold_failed
            and not self.min_absolute_change_failed
            and not self.min_relative_change_failed
        )


class ModelValidationFailedException(MlflowException):
    def __init__(self, message, **kwargs):
        super().__init__(message, error_code=BAD_REQUEST, **kwargs)


def validate_evaluation_results(
    validation_thresholds: dict[str, MetricThreshold],
    candidate_result: EvaluationResult,
    baseline_result: Optional[EvaluationResult] = None,
):
    """
    Validate the evaluation result from one model (candidate) against another
    model (baseline). If the candidate results do not meet the validation
    thresholds, an ModelValidationFailedException will be raised.

    .. note::

        This API is a replacement for the deprecated model validation
        functionality in the :py:func:`mlflow.evaluate` API.

    Args:
        validation_thresholds: A dictionary of metric name to
            :py:class:`mlflow.models.MetricThreshold` used for model validation.
            Each metric name must either be the name of a builtin metric or the
            name of a metric defined in the ``extra_metrics`` parameter.
        candidate_result: The evaluation result of the candidate model.
            Returned by the :py:func:`mlflow.evaluate` API.
        baseline_result: The evaluation result of the baseline model.
            Returned by the :py:func:`mlflow.evaluate` API.
            If set to None, the candidate model result will be
            compared against the threshold values directly.

    Code Example:

        .. code-block:: python
            :caption: Example of Model Validation

            import mlflow
            from mlflow.models import MetricThreshold

            thresholds = {
                "accuracy_score": MetricThreshold(
                    # accuracy should be >=0.8
                    threshold=0.8,
                    # accuracy should be at least 5 percent greater than baseline model accuracy
                    min_absolute_change=0.05,
                    # accuracy should be at least 0.05 greater than baseline model accuracy
                    min_relative_change=0.05,
                    greater_is_better=True,
                ),
            }

            # Get evaluation results for the candidate model
            candidate_result = mlflow.evaluate(
                model="<YOUR_CANDIDATE_MODEL_URI>",
                data=eval_dataset,
                targets="ground_truth",
                model_type="classifier",
            )

            # Get evaluation results for the baseline model
            baseline_result = mlflow.evaluate(
                model="<YOUR_BASELINE_MODEL_URI>",
                data=eval_dataset,
                targets="ground_truth",
                model_type="classifier",
            )

            # Validate the results
            mlflow.validate_evaluation_results(
                thresholds,
                candidate_result,
                baseline_result,
            )

        See `the Model Validation documentation
        <../../models/index.html#performing-model-validation>`_ for more details.
    """
    try:
        assert type(validation_thresholds) is dict
        for key in validation_thresholds.keys():
            assert type(key) is str
        for threshold in validation_thresholds.values():
            assert isinstance(threshold, MetricThreshold)
    except AssertionError:
        raise MlflowException(
            message="The validation thresholds argument must be a dictionary that maps strings "
            "to MetricThreshold objects.",
            error_code=INVALID_PARAMETER_VALUE,
        )

    _logger.info("Validating candidate model metrics against baseline")
    _validate(
        validation_thresholds,
        candidate_result.metrics,
        baseline_result.metrics if baseline_result else {},
    )
    _logger.info("Model validation passed!")


def _validate(
    validation_thresholds: dict[str, MetricThreshold],
    candidate_metrics: dict[str, float],
    baseline_metrics: dict[str, float],
):
    """
    Validate the model based on validation_thresholds by metrics value and
    metrics comparison between candidate model's metrics (candidate_metrics) and
    baseline model's metrics (baseline_metrics).

    Args:
        validation_thresholds: A dictionary from metric_name to MetricThreshold.
        candidate_metrics: The metric evaluation result of the candidate model.
        baseline_metrics: The metric evaluation result of the baseline model.

    Raises:
        If the validation does not pass, raise an MlflowException with detail failure message.
    """
    validation_results = {
        metric_name: _MetricValidationResult(
            metric_name,
            candidate_metrics.get(metric_name),
            threshold,
            baseline_metrics.get(metric_name),
        )
        for (metric_name, threshold) in validation_thresholds.items()
    }

    for metric_name, metric_threshold in validation_thresholds.items():
        validation_result = validation_results[metric_name]

        if metric_name not in candidate_metrics:
            validation_result.missing_candidate = True
            continue

        candidate_metric_value = candidate_metrics[metric_name]
        baseline_metric_value = baseline_metrics[metric_name] if baseline_metrics else None

        # If metric is greater is better, >= is used, otherwise <= is used
        # for thresholding metric value and model comparison
        comparator_fn = operator.__ge__ if metric_threshold.greater_is_better else operator.__le__
        operator_fn = operator.add if metric_threshold.greater_is_better else operator.sub

        if metric_threshold.threshold is not None:
            # metric threshold fails
            # - if not (metric_value >= threshold) for greater is better
            # - if not (metric_value <= threshold) for lower is better
            validation_result.threshold_failed = not comparator_fn(
                candidate_metric_value, metric_threshold.threshold
            )

        if (
            metric_threshold.min_relative_change or metric_threshold.min_absolute_change
        ) and metric_name not in baseline_metrics:
            validation_result.missing_baseline = True
            continue

        if metric_threshold.min_absolute_change is not None:
            # metric comparison absolute change fails
            # - if not (metric_value >= baseline + min_absolute_change) for greater is better
            # - if not (metric_value <= baseline - min_absolute_change) for lower is better
            validation_result.min_absolute_change_failed = not comparator_fn(
                Decimal(candidate_metric_value),
                Decimal(operator_fn(baseline_metric_value, metric_threshold.min_absolute_change)),
            )

        if metric_threshold.min_relative_change is not None:
            # If baseline metric value equals 0, fallback to simple comparison check
            if baseline_metric_value == 0:
                _logger.warning(
                    f"Cannot perform relative model comparison for metric {metric_name} as "
                    "baseline metric value is 0. Falling back to simple comparison: verifying "
                    "that candidate metric value is better than the baseline metric value."
                )
                validation_result.min_relative_change_failed = not comparator_fn(
                    Decimal(candidate_metric_value),
                    Decimal(operator_fn(baseline_metric_value, 1e-10)),
                )
                continue
            # metric comparison relative change fails
            # - if (metric_value - baseline) / baseline < min_relative_change for greater is better
            # - if (baseline - metric_value) / baseline < min_relative_change for lower is better
            if metric_threshold.greater_is_better:
                relative_change = (
                    candidate_metric_value - baseline_metric_value
                ) / baseline_metric_value
            else:
                relative_change = (
                    baseline_metric_value - candidate_metric_value
                ) / baseline_metric_value
            validation_result.min_relative_change_failed = (
                relative_change < metric_threshold.min_relative_change
            )

    failure_messages = []

    for metric_validation_result in validation_results.values():
        if metric_validation_result.is_success():
            continue
        failure_messages.append(str(metric_validation_result))

    if not failure_messages:
        return

    raise ModelValidationFailedException(message=os.linesep.join(failure_messages))



from __future__ import annotations

import json
import logging
from typing import Any, Optional

import agents.tracing as oai
from agents import add_trace_processor
from agents._run_impl import TraceCtxManager
from agents.tracing.setup import GLOBAL_TRACE_PROVIDER
from pydantic import BaseModel

from mlflow.entities.span import LiveSpan, SpanType
from mlflow.entities.span_event import SpanEvent
from mlflow.entities.span_status import SpanStatus, SpanStatusCode
from mlflow.openai import FLAVOR_NAME
from mlflow.tracing.constant import SpanAttributeKey
from mlflow.tracing.fluent import start_span_no_context
from mlflow.types.chat import (
    ChatMessage,
    ChatTool,
    Function,
    FunctionToolDefinition,
    TextContentPart,
    ToolCall,
)
from mlflow.utils.autologging_utils.safety import safe_patch

_logger = logging.getLogger(__name__)


class OpenAISpanType:
    """
    https://github.com/openai/openai-agents-python/blob/ca8e8bed5d0f33e8a0bc3eabd5f1b0a183e73765/src/agents/tracing/span_data.py#L11
    """

    AGENT = "agent"
    FUNCTION = "function"
    GENERATION = "generation"
    RESPONSE = "response"
    HANDOFF = "handoff"
    CUSTOM = "custom"
    GUARDRAIL = "guardrail"


_SPAN_TYPE_MAP = {
    OpenAISpanType.AGENT: SpanType.AGENT,
    OpenAISpanType.FUNCTION: SpanType.TOOL,
    OpenAISpanType.GENERATION: SpanType.CHAT_MODEL,
    OpenAISpanType.RESPONSE: SpanType.CHAT_MODEL,
    OpenAISpanType.GUARDRAIL: SpanType.TOOL,
    # Default to chain type
}


def add_mlflow_trace_processor():
    processors = GLOBAL_TRACE_PROVIDER._multi_processor._processors

    if any(isinstance(p, MlflowOpenAgentTracingProcessor) for p in processors):
        return

    add_trace_processor(MlflowOpenAgentTracingProcessor())


def remove_mlflow_trace_processor():
    processors = GLOBAL_TRACE_PROVIDER._multi_processor._processors
    non_mlflow_processors = [
        p for p in processors if not isinstance(p, MlflowOpenAgentTracingProcessor)
    ]
    GLOBAL_TRACE_PROVIDER._multi_processor._processors = non_mlflow_processors


class MlflowOpenAgentTracingProcessor(oai.TracingProcessor):
    def __init__(
        self,
        project_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self._span_id_to_mlflow_span: dict[str, LiveSpan] = {}
        self._project_name = project_name

        # Patch TraceCtxManager to handle exceptions from the agent properly
        # The original implementation does not propagate exception to the root span,
        # resulting in the trace to have status OK even if there is an exception.
        def _patched_exit(original, instance, exc_type, exc_val, exc_tb):
            try:
                if exc_val and instance.trace:
                    span = self._span_id_to_mlflow_span.get(instance.trace.trace_id)
                    span.add_event(SpanEvent.from_exception(exc_val))
                    span.set_status(SpanStatusCode.ERROR)
            except Exception:
                _logger.debug("Failed to handle exception in MLflow trace", exc_info=True)

            return original(instance, exc_type, exc_val, exc_tb)

        safe_patch(
            FLAVOR_NAME,
            TraceCtxManager,
            "__exit__",
            _patched_exit,
        )

    def on_trace_start(self, trace: oai.Trace) -> None:
        try:
            mlflow_span = start_span_no_context(
                name=trace.name,
                span_type=SpanType.AGENT,
                # TODO: Trace object doesn't contain input/output. Can we get it somehow?
                inputs="",
                attributes=trace.metadata,
            )
            # NB: Trace ID has different prefix as span ID so will not conflict
            self._span_id_to_mlflow_span[trace.trace_id] = mlflow_span

            if trace.group_id:
                # Group ID is used for grouping multiple agent executions together
                mlflow_span.set_tag("group_id", trace.group_id)

            original_exit = trace.__exit__

            # Patch __exit__ method to handle exception properly
            def _patched_exit(self, exc_type, exc_val, exc_tb):
                if exc_val:
                    mlflow_span.add_event(SpanEvent.from_exception(exc_val))
                    mlflow_span.set_status(SpanStatusCode.ERROR)

                original_exit(exc_type, exc_val, exc_tb)

            safe_patch(
                FLAVOR_NAME,
                trace.__class__,
                "__exit__",
                _patched_exit,
            )

        except Exception:
            _logger.debug("Failed to start MLflow trace", exc_info=True)

    def on_trace_end(self, trace: oai.Trace) -> None:
        try:
            mlflow_span = self._span_id_to_mlflow_span.pop(trace.trace_id, None)
            mlflow_span.end(status=mlflow_span.status, outputs="")
        except Exception:
            _logger.debug("Failed to end MLflow trace", exc_info=True)

    def on_span_start(self, span: oai.Span[Any]) -> None:
        try:
            parent_mlflow_span = self._span_id_to_mlflow_span.get(span.parent_id)

            # Parent might be a trace
            if not parent_mlflow_span:
                parent_mlflow_span = self._span_id_to_mlflow_span.get(span.trace_id)

            inputs, _, attributes = _parse_span_data(span.span_data)

            mlflow_span = start_span_no_context(
                name=_get_span_name(span.span_data),
                span_type=_SPAN_TYPE_MAP.get(span.span_data.type, SpanType.CHAIN),
                parent_span=parent_mlflow_span,
                inputs=inputs,
                attributes=attributes,
            )
            self._span_id_to_mlflow_span[span.span_id] = mlflow_span
        except Exception:
            _logger.debug("Failed to start MLflow span", exc_info=True)

    def on_span_end(self, span: oai.Span[Any]) -> None:
        try:
            # parsed_span_data = parse_spandata(span.span_data)
            mlflow_span = self._span_id_to_mlflow_span.pop(span.span_id, None)

            inputs, outputs, attributes = _parse_span_data(span.span_data)

            mlflow_span.set_inputs(inputs)
            mlflow_span.set_outputs(outputs)
            mlflow_span.set_attributes(attributes)

            if span.error:
                status = SpanStatus(
                    status_code=SpanStatusCode.ERROR,
                    description=span.error["message"],
                )
                mlflow_span.add_event(
                    SpanEvent(
                        name="exception",
                        attributes={
                            "exception.message": span.error["message"],
                            "exception.type": "",
                            "exception.stacktrace": json.dumps(span.error["data"]),
                        },
                    )
                )
            else:
                status = SpanStatusCode.OK

            mlflow_span.end(status=status)
        except Exception:
            _logger.debug("Failed to end MLflow span", exc_info=True)

    def force_flush(self) -> None:
        # MLflow doesn't need flush but this method is required by the interface
        pass

    def shutdown(self) -> None:
        self.force_flush()


def _get_span_name(span_data: oai.SpanData) -> str:
    if hasattr(span_data, "name"):
        return span_data.name
    elif isinstance(span_data, oai.GenerationSpanData):
        return "Generation"
    elif isinstance(span_data, oai.ResponseSpanData):
        return "Response"
    elif isinstance(span_data, oai.HandoffSpanData):
        return "Handoff"
    else:
        return "Unknown"


def _parse_span_data(span_data: oai.SpanData) -> tuple[Any, Any, dict[str, Any]]:
    inputs = None
    outputs = None
    attributes = {}

    if span_data.type == OpenAISpanType.AGENT:
        attributes = {
            "handoffs": span_data.handoffs,
            "tools": span_data.tools,
            "output_type": span_data.output_type,
        }
        outputs = {"output_type": span_data.output_type}

    elif span_data.type == OpenAISpanType.FUNCTION:
        try:
            inputs = json.loads(span_data.input)
        except Exception:
            inputs = span_data.input
        outputs = span_data.output

    elif span_data.type == OpenAISpanType.GENERATION:
        inputs = span_data.input
        outputs = span_data.output
        attributes = {
            "model": span_data.model,
            "model_config": span_data.model_config,
            "usage": span_data.usage,
        }

    elif span_data.type == OpenAISpanType.RESPONSE:
        inputs, outputs, attributes = _parse_response_span_data(span_data)

    elif span_data.type == OpenAISpanType.HANDOFF:
        inputs = {"from_agent": span_data.from_agent}
        outputs = {"to_agent": span_data.to_agent}

    elif span_data.type == OpenAISpanType.CUSTOM:
        outputs = span_data.data

    elif span_data.type == OpenAISpanType.GUARDRAIL:
        outputs = {"triggered": span_data.triggered}

    return inputs, outputs, attributes


def _parse_response_span_data(span_data: oai.ResponseSpanData) -> tuple[Any, Any, dict[str, Any]]:
    inputs = span_data.input
    response = span_data.response
    response_dict = response.model_dump() if response else {}
    outputs = response_dict.get("output")
    attributes = {k: v for k, v in response_dict.items() if k != "output"}

    # Extract chat messages
    messages = []
    if response and response.instructions:
        messages.append(ChatMessage(role="system", content=span_data.response.instructions))
    if span_data.input:
        parsed = [_parse_message_like(m) for m in span_data.input]
        messages.extend([m for m in parsed if m is not None])
    if response and response.output:
        parsed = [_parse_message_like(m) for m in span_data.response.output]
        messages.extend(parsed)
    attributes[SpanAttributeKey.CHAT_MESSAGES] = [m.model_dump_compat() for m in messages]

    # Extract chat tools
    chat_tools = []
    for tool in response_dict.get("tools", []):
        try:
            tool = ChatTool(
                type="function",
                function=FunctionToolDefinition(
                    name=tool["name"],
                    description=tool.get("description"),
                    parameters=tool.get("parameters"),
                    strict=tool.get("strict"),
                ),
            )
            chat_tools.append(tool)
        except Exception as e:
            _logger.debug(f"Failed to parse chat tool: {tool}. Error: {e}")

    if chat_tools:
        attributes[SpanAttributeKey.CHAT_TOOLS] = chat_tools

    return inputs, outputs, attributes


def _parse_message_like(message_like: Any) -> Optional[ChatMessage]:
    try:
        return ChatMessage.validate_compat(message_like)
    except Exception:
        pass

    if isinstance(message_like, BaseModel):
        message_like = message_like.model_dump()

    msg_type = message_like["type"]
    if msg_type == "message":
        content = []
        refusal = None
        for content_block in message_like["content"]:
            # Content is a list of either text or refusal https://github.com/openai/openai-python/blob/9dea82fb8cdd06683f9e8033b54cff219789af7f/src/openai/types/responses/response_output_message.py#L13C38-L13C56
            if "text" in content_block:
                content.append(TextContentPart(type="text", text=content_block["text"]))
            elif "refusal" in content_block:
                refusal = content_block["refusal"]
            else:
                _logger.debug(f"Unknown content type in message: {content_block}")
        return ChatMessage(
            role=message_like["role"],
            content=content,
            refusal=refusal,
        )
    elif msg_type == "function_call":
        return ChatMessage(
            role="assistant",
            content="",
            tool_calls=[
                ToolCall(
                    id=message_like["call_id"],
                    function=Function(
                        name=message_like["name"],
                        arguments=message_like["arguments"],
                    ),
                )
            ],
        )
    elif msg_type == "function_call_output":
        return ChatMessage(
            role="tool",
            content=message_like["output"],
            tool_call_id=message_like["call_id"],
        )

    # Ignore unknown message types.
    # Response API supports the following additional message types, which is not
    # supported by our chat standard schema yet:
    # https://github.com/openai/openai-python/blob/9dea82fb8cdd06683f9e8033b54cff219789af7f/src/openai/types/responses/response_output_item.py#L16
    # - File search tool call
    # - Web search tool call
    # - Computer tool call
    # - Reasoning
    _logger.debug(f"Unknown message type: {msg_type}")
